# -*- coding: utf-8 -*-
"""Handle transactions endpoints."""
from .apirequest import APIRequest
from ..exceptions import StreamTerminated
from .decorators import dyndoc_insert, endpoint
from .responses.transactions import responses
from types import GeneratorType
from abc import abstractmethod


class Transactions(APIRequest):
    """Transactions - abstract baseclass to handle transaction endpoints."""

    ENDPOINT = ""
    METHOD = "GET"

    @abstractmethod
    @dyndoc_insert(responses)
    def __init__(self, accountID, transactionID=None):
        """Instantiate a Transactions APIRequest instance.

        Parameters
        ----------
        accountID : string (required)
            the id of the account.

        transactionID : string
            the id of the transaction

        """
        endpoint = self.ENDPOINT.format(accountID=accountID,
                                        transactionID=transactionID)
        super(Transactions, self).__init__(endpoint,
                                           method=self.METHOD)


@endpoint("v3/accounts/{accountID}/transactions")
class TransactionList(Transactions):
    """TransactionList.

    Get a list of Transactions pages that satisfy a time-based Transaction
    query.
    """

    @dyndoc_insert(responses)
    def __init__(self, accountID, params=None):
        """Instantiate a TransactionList request.

        Parameters
        ----------
        accountID : string (required)
            id of the account to perform the request on.

        params : dict (optional)
            query params to send, check developer.oanda.com for details.


        Query Params example::

           {_v3_accounts_accountID_transactions_params}

        >>> import oandapyV20
        >>> import oandapyV20.endpoints.transactions as trans
        >>> client = oandapyV20.API(access_token=...)
        >>> r = trans.TransactionList(accountID)  # params optional
        >>> client.request(r)
        >>> print r.response

        Output::

           {_v3_accounts_accountID_transactions_resp}

        """
        super(TransactionList, self).__init__(accountID)
        self.params = params


@endpoint("v3/accounts/{accountID}/transactions/{transactionID}")
class TransactionDetails(Transactions):
    """Get the details of a single Account Transaction."""

    @dyndoc_insert(responses)
    def __init__(self, accountID, transactionID):
        """Instantiate a TransactionDetails request.

        Parameters
        ----------
        accountID : string (required)
            id of the account to perform the request on.

        transactionID : string (required)
            id of the transaction


        >>> import oandapyV20
        >>> import oandapyV20.endpoints.transactions as trans
        >>> client = oandapyV20.API(access_token=...)
        >>> r = trans.TransactionDetails(accountID=..., transactionID=...)
        >>> client.request(r)
        >>> print r.response

        Output::

           {_v3_accounts_transaction_details_resp}

        """
        super(TransactionDetails, self).__init__(accountID, transactionID)


@endpoint("v3/accounts/{accountID}/transactions/idrange")
class TransactionIDRange(Transactions):
    """TransactionIDRange.

    Get a range of Transactions for an Account based on Transaction IDs.
    """

    @dyndoc_insert(responses)
    def __init__(self, accountID, params=None):
        """Instantiate an TransactionIDRange request.

        Parameters
        ----------
        accountID : string (required)
            id of the account to perform the request on.

        params : dict (required)
            query params to send, check developer.oanda.com for details.


        Query Params example::

            {_v3_accounts_transaction_idrange_params}


        >>> import oandapyV20
        >>> import oandapyV20.endpoints.transactions as trans
        >>> client = oandapyV20.API(access_token=...)
        >>> params = {_v3_accounts_transaction_idrange_params}
        >>> r = trans.TransactionIDRange(accountID=..., params=params)
        >>> client.request(r)
        >>> print r.response

        Output::

            {_v3_accounts_transaction_idrange_resp}

        """
        super(TransactionIDRange, self).__init__(accountID)
        self.params = params


@endpoint("v3/accounts/{accountID}/transactions/sinceid")
class TransactionsSinceID(Transactions):
    """TransactionsSinceID.

    Get a range of Transactions for an Account starting at (but not including)
    a provided Transaction ID.
    """

    @dyndoc_insert(responses)
    def __init__(self, accountID, params=None):
        """Instantiate an TransactionsSince request.

        Parameters
        ----------
        accountID : string (required)
            id of the account to perform the request on.

        params : dict (required)
            query params to send, check developer.oanda.com for details.


        Query Params example::

            {_v3_accounts_transaction_sinceid_params}


        >>> import oandapyV20
        >>> import oandapyV20.endpoints.transactions as trans
        >>> client = oandapyV20.API(access_token=...)
        >>> params = {_v3_accounts_transaction_sinceid_params}
        >>> r = trans.TransactionsSinceID(accountID=..., params=params)
        >>> client.request(r)
        >>> print r.response

        Output::

            {_v3_accounts_transaction_sinceid_resp}

        """
        super(TransactionsSinceID, self).__init__(accountID)
        self.params = params


@endpoint("v3/accounts/{accountID}/transactions/stream")
class TransactionsStream(Transactions):
    """TransactionsStream.

    Get a stream of Transactions for an Account starting from when the
    request is made.
    """

    STREAM = True

    @dyndoc_insert(responses)
    def __init__(self, accountID, params=None):
        """Instantiate an TransactionsStream request.

        Performing this request will result in a generator yielding
        transactions.

        Parameters
        ----------
        accountID : string (required)
            id of the account to perform the request on.


        >>> import oandapyV20
        >>> import oandapyV20.endpoints.transactions as trans
        >>> client = oandapyV20.API(access_token=...)
        >>> r = trans.TransactionsStream(accountID=...)
        >>> rv = client.request(r)
        >>> maxrecs = 5
        >>> try:
        >>>     for T in r.response:  # or rv ...
        >>>         print json.dumps(R, indent=4), ","
        >>>         maxrecs -= 1
        >>>         if maxrecs == 0:
        >>>             r.terminate("Got them all")
        >>> except StreamTerminated as e:
        >>>    print("Finished: {{msg}}".format(msg=e))

        Output::

            {_v3_accounts_transactions_stream_ciresp}

            Finished: Got them all

        """
        super(TransactionsStream, self).__init__(accountID)
        self.params = params

    def terminate(self, message=""):
        """terminate the stream.

        Calling this method will stop the generator yielding transaction
        records. A message can be passed optionally.
        """
        if not isinstance(self.response, GeneratorType):
            raise ValueError("request does not contain a stream response")

        self.response.throw(StreamTerminated(message))



from .marketorder import MarketOrderRequest
from .limitorder import LimitOrderRequest
from .mitorder import MITOrderRequest
from .takeprofitorder import TakeProfitOrderRequest
from .stoplossorder import StopLossOrderRequest
from .trailingstoplossorder import TrailingStopLossOrderRequest
from .stoporder import StopOrderRequest
from .positionclose import PositionCloseRequest
from .tradeclose import TradeCloseRequest

from .onfill import (
    TakeProfitDetails,
    StopLossDetails,
    TrailingStopLossDetails
)
from .extensions import ClientExtensions

__all__ = (
    'MarketOrderRequest',
    'LimitOrderRequest',
    'MITOrderRequest',
    'TakeProfitOrderRequest',
    'StopLossOrderRequest',
    'TrailingStopLossOrderRequest',
    'StopOrderRequest',
    'PositionCloseRequest',
    'TradeCloseRequest',
    'TakeProfitDetails',
    'StopLossDetails',
    'TrailingStopLossDetails',
    'ClientExtensions',
)



# -*- coding: utf-8 -*-

from .baserequest import BaseRequest
from oandapyV20.types import Units, PriceValue
from oandapyV20.definitions.orders import (
    OrderType,
    TimeInForce,
    OrderPositionFill)


class LimitOrderRequest(BaseRequest):
    """create a LimitOrderRequest.

    LimitOrderRequest is used to build the body for a LimitOrder.
    The body can be used to pass to the OrderCreate endpoint.
    """

    def __init__(self,
                 instrument,
                 units,
                 price,
                 positionFill=OrderPositionFill.DEFAULT,
                 clientExtensions=None,
                 takeProfitOnFill=None,
                 timeInForce=TimeInForce.GTC,
                 gtdTime=None,
                 stopLossOnFill=None,
                 trailingStopLossOnFill=None,
                 tradeClientExtensions=None):
        """
        Instantiate a LimitOrderRequest.

        Parameters
        ----------

        instrument : string (required)
            the instrument to create the order for

        units: integer (required)
            the number of units. If positive the order results in a LONG
            order. If negative the order results in a SHORT order

        price: float (required)
            the price indicating the limit.

        Example
        -------

        >>> import json
        >>> from oandapyV20 import API
        >>> import oandapyV20.endpoints.orders as orders
        >>> from oandapyV20.contrib.requests import LimitOrderRequest
        >>>
        >>> accountID = "..."
        >>> client = API(access_token=...)
        >>> ordr = LimitOrderRequest(instrument="EUR_USD",
        ...                          units=10000, price=1.08)
        >>> print(json.dumps(ordr.data, indent=4))
        {
            "order": {
                "timeInForce": "GTC",
                "instrument": "EUR_USD",
                "units": "10000",
                "price": "1.08000",
                "type": "LIMIT",
                "positionFill": "DEFAULT"
            }
        }
        >>> r = orders.orderCreate(accountID, data=ordr.data)
        >>> rv = client.request(r)
        >>>
        """
        super(LimitOrderRequest, self).__init__()

        # by default for a LIMIT order
        self._data.update({"type": OrderType.LIMIT})
        self._data.update({"timeInForce": timeInForce})

        # required
        self._data.update({"instrument": instrument})
        self._data.update({"units": Units(units).value})
        self._data.update({"price": PriceValue(price).value})

        # optional, but required if timeInForce.GTD
        self._data.update({"gtdTime": gtdTime})
        if timeInForce == TimeInForce.GTD and not gtdTime:
            raise ValueError("gtdTime missing")

        # optional
        self._data.update({"positionFill": positionFill})
        self._data.update({"clientExtensions": clientExtensions})
        self._data.update({"takeProfitOnFill": takeProfitOnFill})
        self._data.update({"stopLossOnFill": stopLossOnFill})
        self._data.update({"trailingStopLossOnFill": trailingStopLossOnFill})
        self._data.update({"tradeClientExtensions": tradeClientExtensions})

    @property
    def data(self):
        """data property.

        return the JSON order body
        """
        return dict({"order": super(LimitOrderRequest, self).data})



# -*- coding: utf-8 -*-

from .baserequest import BaseRequest
from oandapyV20.types import Units, PriceValue
from oandapyV20.definitions.orders import (
    OrderType,
    TimeInForce,
    OrderPositionFill)


class MarketOrderRequest(BaseRequest):
    """create a MarketOrderRequest.

    MarketOrderRequest is used to build the body for a MarketOrder.
    The body can be used to pass to the OrderCreate endpoint.
    """

    def __init__(self,
                 instrument,
                 units,
                 priceBound=None,
                 positionFill=OrderPositionFill.DEFAULT,
                 clientExtensions=None,
                 takeProfitOnFill=None,
                 timeInForce=TimeInForce.FOK,
                 stopLossOnFill=None,
                 trailingStopLossOnFill=None,
                 tradeClientExtensions=None):
        """
        Instantiate a MarketOrderRequest.

        Parameters
        ----------

        instrument : string (required)
            the instrument to create the order for

        units: integer (required)
            the number of units. If positive the order results in a LONG
            order. If negative the order results in a SHORT order


        Example
        -------

        >>> import json
        >>> from oandapyV20 import API
        >>> import oandapyV20.endpoints.orders as orders
        >>> from oandapyV20.contrib.requests import MarketOrderRequest
        >>>
        >>> accountID = "..."
        >>> client = API(access_token=...)
        >>> mo = MarketOrderRequest(instrument="EUR_USD", units=10000)
        >>> print(json.dumps(mo.data, indent=4))
        {
            "order": {
                "type": "MARKET",
                "positionFill": "DEFAULT",
                "instrument": "EUR_USD",
                "timeInForce": "FOK",
                "units": "10000"
            }
        }
        >>> # now we have the order specification, create the order request
        >>> r = orders.OrderCreate(accountID, data=mo.data)
        >>> # perform the request
        >>> rv = client.request(r)
        >>> print(rv)
        >>> print(json.dumps(rv, indent=4))
        {
            "orderFillTransaction": {
                "reason": "MARKET_ORDER",
                "pl": "0.0000",
                "accountBalance": "97864.8813",
                "units": "10000",
                "instrument": "EUR_USD",
                "accountID": "101-004-1435156-001",
                "time": "2016-11-11T19:59:43.253587917Z",
                "type": "ORDER_FILL",
                "id": "2504",
                "financing": "0.0000",
                "tradeOpened": {
                    "tradeID": "2504",
                    "units": "10000"
                },
                "orderID": "2503",
                "userID": 1435156,
                "batchID": "2503",
                "price": "1.08463"
            },
            "lastTransactionID": "2504",
            "relatedTransactionIDs": [
                "2503",
                "2504"
            ],
            "orderCreateTransaction": {
                "type": "MARKET_ORDER",
                "reason": "CLIENT_ORDER",
                "id": "2503",
                "timeInForce": "FOK",
                "units": "10000",
                "time": "2016-11-11T19:59:43.253587917Z",
                "positionFill": "DEFAULT",
                "accountID": "101-004-1435156-001",
                "instrument": "EUR_USD",
                "batchID": "2503",
                "userID": 1435156
            }
        }
        >>>
        """
        super(MarketOrderRequest, self).__init__()

        # allowed: FOK/IOC
        if timeInForce not in [TimeInForce.FOK,
                               TimeInForce.IOC]:
            raise ValueError("timeInForce: {}".format(timeInForce))

        # by default for a MARKET order
        self._data.update({"type": OrderType.MARKET})
        self._data.update({"timeInForce": timeInForce})

        # required
        self._data.update({"instrument": instrument})
        self._data.update({"units": Units(units).value})

        # optional
        if priceBound:
            self._data.update({"priceBound": PriceValue(priceBound).value})

        if not hasattr(OrderPositionFill, positionFill):
            raise ValueError("positionFill {}".format(positionFill))

        self._data.update({"positionFill": positionFill})

        self._data.update({"clientExtensions": clientExtensions})
        self._data.update({"takeProfitOnFill": takeProfitOnFill})
        self._data.update({"stopLossOnFill": stopLossOnFill})
        self._data.update({"trailingStopLossOnFill": trailingStopLossOnFill})
        self._data.update({"tradeClientExtensions": tradeClientExtensions})

    @property
    def data(self):
        """data property.

        return the JSON body.
        """
        return dict({"order": super(MarketOrderRequest, self).data})



"""Handling of API requests."""
import six
from abc import ABCMeta, abstractmethod


@six.add_metaclass(ABCMeta)
class APIRequest(object):
    """Base Class for API-request classes."""

    @abstractmethod
    def __init__(self, endpoint, method="GET", expected_status=200):
        """Instantiate an API request.

        Parameters
        ----------
        endpoint : string
            the URL format string

        method : string
            the method for the request. Default: GET.
        """
        self._expected_status = expected_status
        self._status_code = None
        self._response = None

        self._endpoint = endpoint
        self.method = method

    @property
    def expected_status(self):
        return self._expected_status

    @property
    def status_code(self):
        return self._status_code

    @status_code.setter
    def status_code(self, value):
        if value != self._expected_status:
            raise ValueError("{} {} {:d}".format(self, self.method, value))
        self._status_code = value

    @property
    def response(self):
        """response - get the response of the request."""
        return self._response

    @response.setter
    def response(self, value):
        """response - set the response of the request."""
        self._response = value

    def __str__(self):
        """return the endpoint."""
        return self._endpoint



# -*- coding: utf-8 -*-
"""Primitives definitions."""

definitions = {
    "InstrumentType": {
        "CURRENCY": "Currency",
        "CFD": "Contract For Difference",
        "METAL": "Metal"
    },
    "AcceptDatetimeFormat": {
        "UNIX": "Unix timeformat: DateTime fields will be specified "
                "or returned in the “12345678.000000123” format.",
        "RFC3339": "RFC3339 timeformat: DateTime will be specified "
                "or returned in “YYYY-MM-DDTHH:MM:SS.nnnnnnnnnZ” format."
    },
    "Direction": {
       "LONG": "A long Order is used to to buy units of an Instrument. A "
               "Trade is long when it has bought units of an Instrument.",
       "SHORT": "A short Order is used to to sell units of an Instrument. A "
                "Trade is short when it has sold units of an Instrument"
    }
}



from .types import (
    AccountID,
    DateTime,
    OrderID,
    TradeID,
    AccountUnits,
    PriceValue,
    Units,
    ClientID,
    ClientTag,
    ClientComment,
    OrderIdentifier,
    OrderSpecifier
)

__all__ = (
    'AccountID',
    'DateTime',
    'OrderID',
    'TradeID',
    'AccountUnits',
    'PriceValue',
    'Units',
    'ClientID',
    'ClientTag',
    'ClientComment',
    'OrderIdentifier',
    'OrderSpecifier'
)



"""Handle forexlabs endpoints."""
from .apirequest import APIRequest
from .decorators import dyndoc_insert, endpoint
from .responses.forexlabs import responses
from abc import abstractmethod


class ForexLabs(APIRequest):
    """ForexLabs - abstractbase class to handle the 'forexlabs' endpoints."""

    ENDPOINT = ""
    METHOD = "GET"

    @abstractmethod
    @dyndoc_insert(responses)
    def __init__(self):
        """Instantiate a ForexLabs APIRequest instance."""
        endpoint = self.ENDPOINT.format()
        super(ForexLabs, self).__init__(endpoint,
                                        method=self.METHOD)


@endpoint("labs/v1/calendar")
class Calendar(ForexLabs):
    """Calendar.

    Get calendar information.
    """

    @dyndoc_insert(responses)
    def __init__(self, params):
        """Instantiate a Calendar request.

        Parameters
        ----------
        params : dict (required)
            query params to send, check developer.oanda.com for details.


        >>> import oandapyV20
        >>> import oandapyV20.endpoints.forexlabs as labs
        >>> accountID = ...
        >>> client = oandapyV20.API(access_token=...)
        >>> params = {_v3_forexlabs_calendar_params}
        >>> r = labs.Calendar(params=params)
        >>> client.request(r)
        >>> print(r.response)

        Output::

            {_v3_forexlabs_calendar_resp}

        """
        super(Calendar, self).__init__()
        self.params = params


@endpoint("labs/v1/historical_position_ratios")
class HistoricalPositionRatios(ForexLabs):
    """HistoricalPositionRatios.

    Get the historical positionratios for an instrument.
    """

    @dyndoc_insert(responses)
    def __init__(self, params):
        """Instantiate a HistoricalPositionRatios request.

        Parameters
        ----------
        params : dict (required)
            query params to send, check developer.oanda.com for details.


        >>> import oandapyV20
        >>> import oandapyV20.endpoints.forexlabs as labs
        >>> accountID = ...
        >>> client = oandapyV20.API(access_token=...)
        >>> params = {_v3_forexlabs_histposratios_params}
        >>> r = labs.HistoricalPositionRatios(params=params)
        >>> client.request(r)
        >>> print(r.response)

        Output::

            {_v3_forexlabs_histposratios_resp}

        """
        super(HistoricalPositionRatios, self).__init__()
        self.params = params


@endpoint("labs/v1/spreads")
class Spreads(ForexLabs):
    """Spreads.

    Get the spread information for an instrument.
    """

    @dyndoc_insert(responses)
    def __init__(self, params):
        """Instantiate a Spreads request.

        Parameters
        ----------
        params : dict (required)
            query params to send, check developer.oanda.com for details.


        >>> import oandapyV20
        >>> import oandapyV20.endpoints.forexlabs as labs
        >>> accountID = ...
        >>> client = oandapyV20.API(access_token=...)
        >>> params = {_v3_forexlabs_spreads_params}
        >>> r = labs.Spreads(params=params)
        >>> client.request(r)
        >>> print(r.response)

        Output::

            {_v3_forexlabs_spreads_resp}

        """
        super(Spreads, self).__init__()
        self.params = params


@endpoint("labs/v1/commitments_of_traders")
class CommitmentsOfTraders(ForexLabs):
    """CommitmentsOfTraders.

    Get the 'commitments of traders' information for an instrument.
    """

    @dyndoc_insert(responses)
    def __init__(self, params):
        """Instantiate a CommitmentsOfTraders request.

        Parameters
        ----------
        params : dict (required)
            query params to send, check developer.oanda.com for details.


        >>> import oandapyV20
        >>> import oandapyV20.endpoints.forexlabs as labs
        >>> accountID = ...
        >>> client = oandapyV20.API(access_token=...)
        >>> params = {_v3_forexlabs_commoftrad_params}
        >>> r = labs.CommitmentOfTraders(params=params)
        >>> client.request(r)
        >>> print(r.response)

        Output::

            {_v3_forexlabs_commoftrad_resp}

        """
        super(CommitmentsOfTraders, self).__init__()
        self.params = params


@endpoint("labs/v1/orderbook_data")
class OrderbookData(ForexLabs):
    """OrderbookData.

    Get the 'orderbook data' for an instrument.
    """

    @dyndoc_insert(responses)
    def __init__(self, params):
        """Instantiate an OrderbookData request.

        Parameters
        ----------
        params : dict (required)
            query params to send, check developer.oanda.com for details.


        >>> import oandapyV20
        >>> import oandapyV20.endpoints.forexlabs as labs
        >>> accountID = ...
        >>> client = oandapyV20.API(access_token=...)
        >>> params = {_v3_forexlabs_orderbookdata_params}
        >>> r = labs.CommitmentOfTraders(params=params)
        >>> client.request(r)
        >>> print(r.response)

        Output::

            {_v3_forexlabs_orderbookdata_resp}

        """
        super(OrderbookData, self).__init__()
        self.params = params


@endpoint("labs/v1/signal/autochartist")
class Autochartist(ForexLabs):
    """Autochartist.

    Get the 'autochartist data'.
    """

    @dyndoc_insert(responses)
    def __init__(self, params=None):
        """Instantiate an Autochartist request.

        Parameters
        ----------
        params : dict (optional)
            query params to send, check developer.oanda.com for details.


        >>> import oandapyV20
        >>> import oandapyV20.endpoints.forexlabs as labs
        >>> accountID = ...
        >>> client = oandapyV20.API(access_token=...)
        >>> params = {_v3_forexlabs_autochartist_params}
        >>> r = labs.Autochartist(params=params)
        >>> client.request(r)
        >>> print(r.response)

        Output::

            {_v3_forexlabs_autochartist_resp}

        """
        super(Autochartist, self).__init__()
        self.params = params



# -*- coding: utf-8 -*-
"""Handle trades endpoints."""
from .apirequest import APIRequest
from .decorators import dyndoc_insert, endpoint
from .responses.trades import responses
from abc import abstractmethod


class Trades(APIRequest):
    """Trades - abstract baseclass to handle the trades endpoints."""

    ENDPOINT = ""
    METHOD = "GET"

    @abstractmethod
    @dyndoc_insert(responses)
    def __init__(self, accountID, tradeID=None):
        """Instantiate a Trades APIRequest instance.

        Parameters
        ----------
        accountID : string
            the account_id of the account.

        tradeID : string
            ID of the trade

        """
        endpoint = self.ENDPOINT.format(accountID=accountID, tradeID=tradeID)
        super(Trades, self).__init__(endpoint, method=self.METHOD)


@endpoint("v3/accounts/{accountID}/trades")
class TradesList(Trades):
    """Get a list of trades for an Account."""

    @dyndoc_insert(responses)
    def __init__(self, accountID, params=None):
        """Instantiate a TradesList request.

        Parameters
        ----------
        accountID : string (required)
            id of the account to perform the request on.

        params : dict (optional)
            query params to send, check developer.oanda.com for details.


        Query Params example::

            {_v3_accounts_accountID_trades_params}


        >>> import oandapyV20
        >>> import oandapyV20.endpoints.trades as trades
        >>> client = oandapyV20.API(access_token=...)
        >>> params = {_v3_accounts_accountID_trades_params}
        >>> r = trades.TradesList(accountID=..., params=params)
        >>> client.request(r)
        >>> print r.response

        Output::

            {_v3_accounts_accountID_trades_resp}

        """
        super(TradesList, self).__init__(accountID)
        self.params = params


@endpoint("v3/accounts/{accountID}/openTrades")
class OpenTrades(Trades):
    """Get the list of open Trades for an Account."""

    @dyndoc_insert(responses)
    def __init__(self, accountID):
        """Instantiate an OpenTrades request.

        Parameters
        ----------
        accountID : string (required)
            id of the account to perform the request on.


        >>> import oandapyV20
        >>> import oandapyV20.endpoints.trades as trades
        >>> client = oandapyV20.API(access_token=...)
        >>> r = trades.OpenTrades(accountID=...)
        >>> client.request(r)
        >>> print r.response


        Output::

            {_v3_accounts_accountID_opentrades_resp}

        """
        super(OpenTrades, self).__init__(accountID)


@endpoint("v3/accounts/{accountID}/trades/{tradeID}")
class TradeDetails(Trades):
    """Get the details of a specific Trade in an Account."""

    @dyndoc_insert(responses)
    def __init__(self, accountID, tradeID):
        """Instantiate a TradeDetails request.

        Parameters
        ----------
        accountID : string (required)
            id of the account to perform the request on.

        tradeID : string (required)
            id of the trade.


        >>> import oandapyV20
        >>> import oandapyV20.endpoints.trades as trades
        >>> client = oandapyV20.API(access_token=...)
        >>> r = accounts.TradeDetails(accountID=..., tradeID=...)
        >>> client.request(r)
        >>> print r.response


        Output::

            {_v3_account_accountID_trades_details_resp}

        """
        super(TradeDetails, self).__init__(accountID, tradeID)


@endpoint("v3/accounts/{accountID}/trades/{tradeID}/close", "PUT")
class TradeClose(Trades):
    """TradeClose.

    Close (partially or fully) a specific open Trade in an Account.
    """

    HEADERS = {"Content-Type": "application/json"}

    @dyndoc_insert(responses)
    def __init__(self, accountID, tradeID, data=None):
        """Instantiate a TradeClose request.

        Parameters
        ----------
        accountID : string (required)
            id of the account to perform the request on.

        tradeID : string (required)
            id of the trade to close.

        data : dict (optional)
            data to send, use this to close a trade partially. Check
            developer.oanda.com for details.


        Data body example::

            {_v3_account_accountID_trades_close_body}


        >>> import oandapyV20
        >>> import oandapyV20.endpoints.trades as trades
        >>> client = oandapyV20.API(access_token=...)
        >>> data = {_v3_account_accountID_trades_close_body}
        >>> r = trades.TradeClose(accountID=..., data=data)
        >>> client.request(r)
        >>> print r.response

        Output::

            {_v3_account_accountID_trades_close_resp}

        """
        super(TradeClose, self).__init__(accountID, tradeID)
        self.data = data


@endpoint("v3/accounts/{accountID}/trades/{tradeID}/clientExtensions", "PUT")
class TradeClientExtensions(Trades):
    """TradeClientExtensions.

    Update the Client Extensions for a Trade. Do not add, update or delete
    the Client Extensions if your account is associated with MT4.
    """

    HEADERS = {"Content-Type": "application/json"}

    @dyndoc_insert(responses)
    def __init__(self, accountID, tradeID, data=None):
        """Instantiate a TradeClientExtensions request.

        Parameters
        ----------
        accountID : string (required)
            id of the account to perform the request on.

        tradeID : string (required)
            id of the trade to update client extensions for.

        data : dict (required)
            clientextension data to send, check developer.oanda.com
            for details.


        Data body example::

            {_v3_account_accountID_trades_cltext_body}


        >>> import oandapyV20
        >>> import oandapyV20.endpoints.trades as trades
        >>> accountID = ...
        >>> tradeID = ...
        >>> client = oandapyV20.API(access_token=...)
        >>> data = {_v3_account_accountID_trades_cltext_body}
        >>> r = trades.TradeClientExtensions(accountID=accountID,
        >>>                                  tradeID=tradeID,
        >>>                                  data=data)
        >>> client.request(r)
        >>> print r.response

        Output::

            {_v3_account_accountID_trades_cltext_resp}

        """
        super(TradeClientExtensions, self).__init__(accountID, tradeID)
        self.data = data


@endpoint("v3/accounts/{accountID}/trades/{tradeID}/orders", "PUT")
class TradeCRCDO(Trades):
    """Trade Create Replace Cancel Dependent Orders."""

    HEADERS = {"Content-Type": "application/json"}

    @dyndoc_insert(responses)
    def __init__(self, accountID, tradeID, data):
        """Instantiate a TradeClientExtensions request.

        Parameters
        ----------
        accountID : string (required)
            id of the account to perform the request on.

        tradeID : string (required)
            id of the trade to update client extensions for.

        data : dict (required)
            clientextension data to send, check developer.oanda.com
            for details.


        Data body example::

            {_v3_account_accountID_trades_crcdo_body}


        >>> import oandapyV20
        >>> import oandapyV20.endpoints.trades as trades
        >>> accountID = ...
        >>> tradeID = ...
        >>> client = oandapyV20.API(access_token=...)
        >>> data = {_v3_account_accountID_trades_crcdo_body}
        >>> r = trades.TradeCRCDO(accountID=accountID,
        >>>                       tradeID=tradeID,
        >>>                       data=data)
        >>> client.request(r)
        >>> print r.response

        Output::

            {_v3_account_accountID_trades_crcdo_resp}

        """
        super(TradeCRCDO, self).__init__(accountID, tradeID)
        self.data = data



import re
import time
from datetime import datetime


def secs2time(e):
    """secs2time - convert epoch to datetime.

    >>> d = secs2time(1497499200)
    >>> d
    datetime.datetime(2017, 6, 15, 4, 0)
    >>> d.strftime("%Y%m%d-%H:%M:%S")
    '20170615-04:00:00'
    """
    w = time.gmtime(e)
    return datetime(*list(w)[0:6])


def granularity_to_time(s):
    """convert a named granularity into seconds.

    get value in seconds for named granularities: M1, M5 ... H1 etc.

    >>> print(granularity_to_time("M5"))
    300
    """
    mfact = {
        'S': 1,
        'M': 60,
        'H': 3600,
        'D': 86400,
    }
    try:
        f, n = re.match("(?P<f>[SMHD])(?:(?P<n>\d+)|)", s).groups()
        n = n if n else 1
        return mfact[f] * int(n)

    except Exception as e:
        raise ValueError(e)



"""Responses.

responses serve both testing purpose aswell as dynamic docstring replacement
"""
responses = {
    "_v3_accounts_accountID_trades": {
        "url": "v3/accounts/{accountID}/trades",
        "params": {
            "instrument": "DE30_EUR,EUR_USD"
        },
        "response": {
            "trades": [
                {
                    "financing": "0.0000",
                    "openTime": "2016-10-28T14:28:05.231759081Z",
                    "price": "10678.3",
                    "unrealizedPL": "25.0000",
                    "realizedPL": "0.0000",
                    "instrument": "DE30_EUR",
                    "state": "OPEN",
                    "initialUnits": "10",
                    "currentUnits": "10",
                    "id": "2315"
                },
                {
                    "financing": "0.0000",
                    "openTime": "2016-10-28T14:27:19.011002322Z",
                    "price": "1.09448",
                    "unrealizedPL": "-0.0933",
                    "realizedPL": "0.0000",
                    "instrument": "EUR_USD",
                    "state": "OPEN",
                    "initialUnits": "100",
                    "currentUnits": "100",
                    "id": "2313"
                }
            ],
            "lastTransactionID": "2315"
        }
    },
    "_v3_accounts_accountID_opentrades": {
        "url": "v3/accounts/{accountID}/openTrades",
        "response": {
            "trades": [
                {
                    "financing": "0.0000",
                    "openTime": "2016-10-28T14:28:05.231759081Z",
                    "price": "10678.3",
                    "unrealizedPL": "136.0000",
                    "realizedPL": "0.0000",
                    "instrument": "DE30_EUR",
                    "state": "OPEN",
                    "initialUnits": "10",
                    "currentUnits": "10",
                    "id": "2315"
                }
            ],
            "lastTransactionID": "2317"
        }
    },
    "_v3_account_accountID_trades_details": {
        # tradeID 2315
        "url": "v3/accounts/{accountID}/trades/{tradeID}",
        "response": {
            "lastTransactionID": "2317",
            "trade": {
                "financing": "0.0000",
                "openTime": "2016-10-28T14:28:05.231759081Z",
                "price": "10678.3",
                "unrealizedPL": "226.0000",
                "realizedPL": "0.0000",
                "instrument": "DE30_EUR",
                "state": "OPEN",
                "initialUnits": "10",
                "currentUnits": "10",
                "id": "2315"
            }
        }
    },
    "_v3_account_accountID_trades_close": {
        "url": "v3/accounts/{accountID}/trades/{tradeID}/close",
        "body": {
            "units": 100
        },
        "response": {
            "orderFillTransaction": {
                "orderID": "2316",
                "financing": "0.0000",
                "instrument": "EUR_USD",
                "price": "1.09289",
                "userID": 1435156,
                "batchID": "2316",
                "accountBalance": "33848.1208",
                "reason": "MARKET_ORDER_TRADE_CLOSE",
                "tradesClosed": [
                    {
                        "units": "-100",
                        "financing": "0.0000",
                        "realizedPL": "-0.1455",
                        "tradeID": "2313"
                    }
                ],
                "time": "2016-10-28T15:11:58.023004583Z",
                "units": "-100",
                "type": "ORDER_FILL",
                "id": "2317",
                "pl": "-0.1455",
                "accountID": "101-004-1435156-001"
            },
            "orderCreateTransaction": {
                "timeInForce": "FOK",
                "positionFill": "REDUCE_ONLY",
                "userID": 1435156,
                "batchID": "2316",
                "instrument": "EUR_USD",
                "reason": "TRADE_CLOSE",
                "tradeClose": {
                    "units": "100",
                    "tradeID": "2313"
                },
                "time": "2016-10-28T15:11:58.023004583Z",
                "units": "-100",
                "type": "MARKET_ORDER",
                "id": "2316",
                "accountID": "101-004-1435156-001"
            },
            "relatedTransactionIDs": [
                "2316",
                "2317"
            ],
            "lastTransactionID": "2317"
        }
    },
    "_v3_account_accountID_trades_cltext": {
        "url": "v3/accounts/{accountID}/trades/{tradeID}/close",
        "body": {
            "clientExtensions": {
                "comment": "myComment",
                "id": "myID2315",
            }
        },
        "response": {
            "tradeClientExtensionsModifyTransaction": {
                "tradeID": "2315",
                "userID": 1435156,
                "batchID": "2319",
                "time": "2016-10-28T20:32:39.356516787Z",
                "tradeClientExtensionsModify": {
                    "comment": "myComment",
                    "id": "myID2315"
                },
                "type": "TRADE_CLIENT_EXTENSIONS_MODIFY",
                "id": "2319",
                "accountID": "101-004-1435156-001"
            },
            "relatedTransactionIDs": [
                "2319"
            ],
            "lastTransactionID": "2319"
        }
    },
    "_v3_account_accountID_trades_crcdo": {
        "url": "v3/accounts/{accountID}/trades/{tradeID}/close",
        "body": {
            "takeProfit": {
                "timeInForce": "GTC",
                "price": "1.05"
            },
            "stopLoss": {
                "timeInForce": "GTC",
                "price": "1.10"
            }
        },
        "response": {
            "stopLossOrderTransaction": {
                "timeInForce": "GTC",
                "triggerCondition": "TRIGGER_DEFAULT",
                "replacesOrderID": "2324",
                "tradeID": "2323",
                "price": "1.10000",
                "userID": 1435156,
                "batchID": "2325",
                "reason": "REPLACEMENT",
                "time": "2016-10-28T21:00:19.978476830Z",
                "cancellingTransactionID": "2326",
                "type": "STOP_LOSS_ORDER",
                "id": "2327",
                "accountID": "101-004-1435156-001"
            },
            "takeProfitOrderTransaction": {
                "timeInForce": "GTC",
                "triggerCondition": "TRIGGER_DEFAULT",
                "tradeID": "2323",
                "price": "1.05000",
                "userID": 1435156,
                "batchID": "2325",
                "reason": "CLIENT_ORDER",
                "time": "2016-10-28T21:00:19.978476830Z",
                "type": "TAKE_PROFIT_ORDER",
                "id": "2325",
                "accountID": "101-004-1435156-001"
            },
            "relatedTransactionIDs": [
                "2325",
                "2326",
                "2327"
            ],
            "lastTransactionID": "2327",
            "stopLossOrderCancelTransaction": {
                "orderID": "2324",
                "replacedByOrderID": "2327",
                "userID": 1435156,
                "batchID": "2325",
                "reason": "CLIENT_REQUEST_REPLACED",
                "time": "2016-10-28T21:00:19.978476830Z",
                "type": "ORDER_CANCEL",
                "id": "2326",
                "accountID": "101-004-1435156-001"
            }
        }
    }
}



# -*- coding: utf-8 -*-
"""tradeclose."""

from .baserequest import BaseRequest


class TradeCloseRequest(BaseRequest):
    """create a TradeCloseRequest.

    TradeCloseRequest is used to build the body to close a trade.
    The body can be used to pass to the TradeClose endpoint.
    """

    def __init__(self, units="ALL"):
        """
        Instantiate a TradeCloseRequest.

        Parameters
        ----------

        units : integer (optional)
            the number of units to close. Default it is set to "ALL".

        Example
        -------

        >>> import json
        >>> from oandapyV20 import API
        >>> import oandapyV20.endpoints.trades as trades
        >>> from oandapyV20.contrib.requests import TradeCloseRequest
        >>>
        >>> accountID = "..."
        >>> client = API(access_token=...)
        >>> ordr = TradeCloseRequest(units=10000)
        >>> print(json.dumps(ordr.data, indent=4))
        {
           "units": "10000"
        }
        >>> # now we have the order specification, create the order request
        >>> r = trades.TradeClose(accountID, tradeID=1234,
        >>>                       data=ordr.data)
        >>> # perform the request
        >>> rv = client.request(r)
        >>> print(rv)
        >>> ...
        """
        super(TradeCloseRequest, self).__init__()

        # by default for a TradeClose no parameters are required
        if units:
            self._data.update(
                {"units":
                 "{:d}".format(int(units)) if units != "ALL" else "ALL"})



# -*- coding: utf-8 -*-
"""types."""

import six
import re
from abc import ABCMeta
import datetime as natdatetime  # native datetime


@six.add_metaclass(ABCMeta)
class OAType(object):
    """baseclass for OANDA types."""

    @property
    def value(self):
        """value property."""
        return self._v


class AccountID(OAType):
    """representation of an AccountID, string value of an Account Identifier.

    Parameters
    ----------

    accountID : string (required)
        the accountID of a v20 account

    Example
    -------

        >>> print AccountID("001-011-5838423-001").value


    A ValueError exception is raised in case of an incorrect value.
    """

    def __init__(self, accountID):
        l = re.match(r"(?P<siteID>\d+)-(?P<divisionID>\d+)"
                     "-(?P<userID>\d+)-(?P<accountNumber>\d+)", accountID)
        if not l:
            msg = "AccountID {} not a valid V20 account".format(accountID)
            raise ValueError(msg)

        self._v = l.groupdict()


class OrderID(OAType):
    """representation of an orderID, string value of an integer.

    Parameters
    ----------

    orderID : integer or string (required)
        the orderID as a positive integer or as a string

    Example
    -------

        >>> print OrderID(1234).value


    A ValueError exception is raised in case of a negative integer value
    """

    def __init__(self, orderID):
        if int(orderID) < 0:
            raise ValueError("OrderID must be a positive integer value")
        self._v = "{:d}".format(int(orderID))


class DateTime(OAType):
    """representation of a DateTime as a RFC 3339 string.

    Parameters
    ----------

    dateTime : string, datetime instance, dict (required)
        the dateTime parameter must be:
         - a valid RFC3339 string representing a date-time, or
         - a dict holding the relevant datetime parts, or
         - a datetime.datetime instance

    The value property is always RFC3339  datetime string

    Fractional seconds are in microseconds. This compatible with
    datetime.datetime.

    Example
    -------

    >>> print DateTime("2014-07-02T04:00:00.000000Z").value
    >>> print DateTime({"year": 2014, "month": 12, "day": 2,
    ...                 "hour": 13, "minute": 48, "second": 12}).value
    >>> from datetime import datetime
    >>> print DateTime(datetime.now()).value


    A ValueError exception is raised in case of an invalid value
    """

    def __init__(self, dateTime):

        def formatDT(dtd):

            _date = natdatetime.datetime(
                int(dtd.get("year")),
                int(dtd.get("month")),
                int(dtd.get("day")),
                int(dtd.get("hour")),
                int(dtd.get("minute")),
                int(dtd.get("second")))

            dt = natdatetime.datetime.strftime(_date, "%Y-%m-%dT%H:%M:%S")

            if "subsecond" in dtd and dtd.get("subsecond") is not None:
                dt = "{}.{:>06d}".format(dt, int(dtd.get("subsecond")))

            return dt+"Z"

        if isinstance(dateTime, str):
            l = re.match(r"(?P<year>\d+)-(?P<month>\d+)-(?P<day>\d+)"
                         "T(?P<hour>\d+):(?P<minute>\d+):(?P<second>\d+)"
                         "(?:.(?P<subsecond>\d{1,6})|)"
                         "Z",
                         dateTime)

            if not l:
                msg = "Invalid RFC 3339 string: {}".format(dateTime)
                raise ValueError(msg)

            # print l.groupdict()
            self._v = formatDT(l.groupdict())

        elif isinstance(dateTime, dict):
            self._v = formatDT(dateTime)

        elif isinstance(dateTime, natdatetime.datetime):
            self._v = formatDT({"year": dateTime.year,
                                "month": dateTime.month,
                                "day": dateTime.day,
                                "hour": dateTime.hour,
                                "minute": dateTime.minute,
                                "second": dateTime.second,
                                "subsecond": dateTime.microsecond})


class TradeID(OAType):
    """representation of a tradeID, string value of an integer.

    Parameters
    ----------

    tradeID : integer or string (required)
        the tradeID as a positive integer or as a string

    Example
    -------

        >>> print TradeID(1234).value


    A ValueError exception is raised in case of a negative integer value
    """

    def __init__(self, tradeID):
        if int(tradeID) < 0:
            raise ValueError("TradeID must be a positive integer value")
        self._v = "{:d}".format(int(tradeID))


class AccountUnits(OAType):
    """representation AccountUnits, string value of a float."""

    def __init__(self, units):
        self._v = "{:.5f}".format(float(units))


class PriceValue(OAType):
    """representation PriceValue, string value of a float."""

    def __init__(self, priceValue):
        self._v = "{:.5f}".format(float(priceValue))


class Units(OAType):
    """representation Units, string value of an integer."""

    def __init__(self, units):
        self._v = "{:d}".format(int(units))


class ClientID(OAType):
    """representation of ClientID, a string value of max 128 chars."""

    def __init__(self, clientID):
        length = len(clientID)
        if not length or length > 128:
            raise ValueError("ClientID: length {}".format(length))

        self._v = clientID


class ClientTag(OAType):
    """representation of ClientTag, a string value of max 128 chars."""

    def __init__(self, clientTag):
        length = len(clientTag)
        if not length or length > 128:
            raise ValueError("ClientTag: length {}".format(length))

        self._v = clientTag


class ClientComment(OAType):
    """representation of ClientComment, a string value of max 128 chars."""

    def __init__(self, clientComment):
        length = len(clientComment)
        if not length or length > 128:
            raise ValueError("ClientComment: length {}".format(length))

        self._v = clientComment


class OrderIdentifier(OAType):
    """representation of the OrderIdentifier object."""

    def __init__(self, orderID, clientID):
        self._v = {
            "orderID": OrderID(orderID).value,
            "clientOrderID": ClientID(clientID).value
        }


class OrderSpecifier(OAType):
    """representation of the OrderSpecifier."""

    def __init__(self, specifier):
        if str(specifier).startswith('@'):
            self._v = ClientID(specifier.lstrip('@')).value
        else:
            self._v = OrderID(specifier).value



"""Responses.

responses serve both testing purpose aswell as dynamic docstring replacement
"""
responses = {
    "_v3_accounts_accountID_positions": {
        "url": "v3/accounts/{accountID}/positions",
        "response": {
            "positions": [
                {
                    "short": {
                        "units": "0",
                        "resettablePL": "-272.6805",
                        "unrealizedPL": "0.0000",
                        "pl": "-272.6805"
                    },
                    "unrealizedPL": "0.0000",
                    "long": {
                        "units": "0",
                        "resettablePL": "0.0000",
                        "unrealizedPL": "0.0000",
                        "pl": "0.0000"
                    },
                    "instrument": "EUR_GBP",
                    "resettablePL": "-272.6805",
                    "pl": "-272.6805"
                },
                {
                    "short": {
                        "unrealizedPL": "870.0000",
                        "units": "-20",
                        "resettablePL": "-13959.3000",
                        "tradeIDs": [
                            "2121",
                            "2123"
                        ],
                        "averagePrice": "10581.5",
                        "pl": "-13959.3000"
                    },
                    "unrealizedPL": "870.0000",
                    "long": {
                        "units": "0",
                        "resettablePL": "404.5000",
                        "unrealizedPL": "0.0000",
                        "pl": "404.5000"
                    },
                    "instrument": "DE30_EUR",
                    "resettablePL": "-13554.8000",
                    "pl": "-13554.8000"
                },
                {
                    "short": {
                        "units": "0",
                        "resettablePL": "0.0000",
                        "unrealizedPL": "0.0000",
                        "pl": "0.0000"
                    },
                    "unrealizedPL": "0.0000",
                    "long": {
                        "units": "0",
                        "resettablePL": "-12.8720",
                        "unrealizedPL": "0.0000",
                        "pl": "-12.8720"
                    },
                    "instrument": "EUR_USD",
                    "resettablePL": "-12.8720",
                    "pl": "-12.8720"
                }
            ],
            "lastTransactionID": "2124"
        }
    },
    "_v3_accounts_accountID_openpositions": {
        "url": "v3/accounts/{accountID}/positions",
        "response": {
            "positions": [
                {
                    "short": {
                        "units": "0",
                        "resettablePL": "-14164.3000",
                        "unrealizedPL": "0.0000",
                        "pl": "-14164.3000"
                    },
                    "unrealizedPL": "-284.0000",
                    "long": {
                        "unrealizedPL": "-284.0000",
                        "units": "10",
                        "resettablePL": "404.5000",
                        "tradeIDs": [
                            "2315"
                        ],
                        "averagePrice": "10678.3",
                        "pl": "404.5000"
                    },
                    "instrument": "DE30_EUR",
                    "resettablePL": "-13759.8000",
                    "pl": "-13759.8000"
                },
                {
                    "short": {
                        "unrealizedPL": "-0.0738",
                        "units": "-100",
                        "resettablePL": "0.0000",
                        "tradeIDs": [
                            "2323"
                        ],
                        "averagePrice": "1.09843",
                        "pl": "0.0000"
                    },
                    "unrealizedPL": "-0.0738",
                    "long": {
                        "units": "0",
                        "resettablePL": "-44.6272",
                        "unrealizedPL": "0.0000",
                        "pl": "-44.6272"
                    },
                    "instrument": "EUR_USD",
                    "resettablePL": "-44.6272",
                    "pl": "-44.6272"
                }
            ],
            "lastTransactionID": "2327"
        }
    },
    "_v3_accounts_accountID_positiondetails": {
        "url": "v3/accounts/{accountID}/positions/{instrument}",
        "response": {
            "position": {
                "short": {
                    "unrealizedPL": "-0.0738",
                    "units": "-100",
                    "resettablePL": "0.0000",
                    "tradeIDs": [
                        "2323"
                    ],
                    "averagePrice": "1.09843",
                    "pl": "0.0000"
                },
                "unrealizedPL": "-0.0738",
                "long": {
                    "units": "0",
                    "resettablePL": "-44.6272",
                    "unrealizedPL": "0.0000",
                    "pl": "-44.6272"
                },
                "instrument": "EUR_USD",
                "resettablePL": "-44.6272",
                "pl": "-44.6272"
            },
            "lastTransactionID": "2327"
        }
    },
    "_v3_accounts_accountID_position_close": {
        "url": "v3/accounts/{accountID}/positions/{instrument}/close",
        "body": {
            "longUnits": "ALL"
        },
        "response": {
            "lastTransactionID": "6391",
            "longOrderCreateTransaction": {
                "accountID": "<ACCOUNT>",
                "batchID": "6390",
                "id": "6390",
                "instrument": "EUR_USD",
                "longPositionCloseout": {
                    "instrument": "EUR_USD",
                    "units": "ALL"
                },
                "positionFill": "REDUCE_ONLY",
                "reason": "POSITION_CLOSEOUT",
                "time": "2016-06-22T18:41:35.034041665Z",
                "timeInForce": "FOK",
                "type": "MARKET_ORDER",
                "units": "-251",
                "userID": "<USERID>"
            },
            "longOrderFillTransaction": {
                "accountBalance": "43650.69807",
                "accountID": "<ACCOUNT>",
                "batchID": "6390",
                "financing": "0.00000",
                "id": "6391",
                "instrument": "EUR_USD",
                "orderID": "6390",
                "pl": "-0.03370",
                "price": "1.13018",
                "reason": "MARKET_ORDER_POSITION_CLOSEOUT",
                "time": "2016-06-22T18:41:35.034041665Z",
                "tradesClosed": [
                    {
                        "financing": "0.00000",
                        "realizedPL": "-0.00013",
                        "tradeID": "6383",
                        "units": "-1"
                    },
                    {
                        "financing": "0.00000",
                        "realizedPL": "-0.03357",
                        "tradeID": "6385",
                        "units": "-250"
                    }
                ],
                "type": "ORDER_FILL",
                "units": "-251",
                "userID": "<USERID>"
            },
            "relatedTransactionIDs": [
                "6390",
                "6391"
            ]
        }
    }
}



# -*- coding: utf-8 -*-

from .baserequest import BaseRequest
from oandapyV20.types import ClientID, ClientTag, ClientComment


class ClientExtensions(BaseRequest):
    """Representation of the ClientExtensions."""

    def __init__(self,
                 clientID=None,
                 clientTag=None,
                 clientComment=None):
        """Instantiate ClientExtensions.

        Parameters
        ----------

        clientID : clientID (required)
            the clientID

        clientTag : clientTag (required)
            the clientTag

        clientComment : clientComment (required)
            the clientComment


        Example
        -------

            >>> import json
            >>> from oandapyV20 import API
            >>> import oandapyV20.endpoints.orders as orders
            >>> from oandapyV20.contrib.requests import (
            ...     MarketOrderRequest, TakeProfitDetails, ClientExtensions)
            >>>
            >>> accountID = "..."
            >>> client = API(access_token=...)
            >>> # at time of writing EUR_USD = 1.0740
            >>> # let us take profit at 1.10, GoodTillCancel (default)
            >>> # add clientExtensions to it also
            >>> takeProfitOnFillOrder = TakeProfitDetails(
            ...     price=1.10,
            ...     clientExtensions=ClientExtensions(clientTag="mytag").data)
            >>> print(takeProfitOnFillOrder.data)
            {
                'timeInForce': 'GTC',
                'price": '1.10000',
                'clientExtensions': {'tag': 'mytag'}
            }
            >>> ordr = MarketOrderRequest(
            ...     instrument="EUR_USD",
            ...     units=10000,
            ...     takeProfitOnFill=takeProfitOnFillOrder.data
            ... )
            >>> # or as shortcut ...
            >>> #   takeProfitOnFill=TakeProfitDetails(price=1.10).data
            >>> print(json.dumps(ordr.data, indent=4))
            >>> r = orders.OrderCreate(accountID, data=ordr.data)
            >>> rv = client.request(r)
            >>> ...
        """
        super(ClientExtensions, self).__init__()
        if not (clientID or clientTag or clientComment):
            raise ValueError("clientID, clientTag, clientComment required")

        if clientID:
            self._data.update({"id": ClientID(clientID).value})

        if clientTag:
            self._data.update({"tag": ClientTag(clientTag).value})

        if clientComment:
            self._data.update({"comment": ClientComment(clientComment).value})



# -*- coding: utf-8 -*-

from .baserequest import BaseRequest
from oandapyV20.types import TradeID, PriceValue
from oandapyV20.definitions.orders import TimeInForce, OrderType


class TrailingStopLossOrderRequest(BaseRequest):
    """create a TrailingStopLossOrderRequest.

    TrailingStopLossOrderRequest is used to build the body for a
    TrailingStopLossOrder. The body can be used to pass to the
    OrderCreate endpoint.
    """

    def __init__(self,
                 tradeID,
                 distance,
                 clientTradeID=None,
                 timeInForce=TimeInForce.GTC,
                 gtdTime=None,
                 clientExtensions=None):
        """
        Instantiate a TrailingStopLossOrderRequest.

        Parameters
        ----------

        tradeID : string (required)
            the tradeID of an existing trade

        distance : float (required)
            the price distance

        Example
        -------

        >>> import json
        >>> from oandapyV20 import API
        >>> import oandapyV20.endpoints.orders as orders
        >>> from oandapyV20.contrib.requests import TrailingStopLossOrderRequest
        >>>
        >>> accountID = "..."
        >>> client = API(access_token=...)
        >>> ordr = TrailingStopLossOrderRequest(tradeID="1234", distance=20)
        >>> print(json.dumps(ordr.data, indent=4))
        {
            "order": {
                "type": "TRAILING_STOP_LOSS",
                "tradeID": "1234",
                "timeInForce": "GTC",
                "distance": "20.00000"
            }
        }
        >>> # now we have the order specification, create the order request
        >>> r = orders.OrderCreate(accountID, data=ordr.data)
        >>> # perform the request
        >>> rv = client.request(r)
        >>> print(json.dumps(rv, indent=4))
        >>> ...
        """
        super(TrailingStopLossOrderRequest, self).__init__()

        # allowed: GTC/GFD/GTD
        if timeInForce not in [TimeInForce.GTC,
                               TimeInForce.GTD,
                               TimeInForce.GFD]:
            raise ValueError("timeInForce: {}".format(timeInForce))

        # by default for a TRAILING_STOP_LOSS order
        self._data.update({"type": OrderType.TRAILING_STOP_LOSS})

        # required
        self._data.update({"tradeID": TradeID(tradeID).value})
        self._data.update({"distance": PriceValue(distance).value})

        # optional
        self._data.update({"clientExtensions": clientExtensions})
        self._data.update({"timeInForce": timeInForce})
        self._data.update({"gtdTime": gtdTime})

        if timeInForce == TimeInForce.GTD and not gtdTime:
            raise ValueError("gtdTime missing")

    @property
    def data(self):
        """data property.

        return the JSON body.
        """
        return dict({"order": super(TrailingStopLossOrderRequest, self).data})



# -*- coding: utf-8 -*-
"""Account Definitions."""

definitions = {
    "AccountFinancingMode": {
        "NO_FINANCING": "No financing is paid/charged for open Trades "
                        "in the Account",
        "SECOND_BY_SECOND": "Second-by-second financing is paid/charged "
                            "for open Trades in the Account, both daily "
                            "and when the the Trade is closed",
        "DAILY": "A full day’s worth of financing is paid/charged for "
                 "open Trades in the Account daily at 5pm New York time"
    },
    "PositionAggregationMode": {
        "ABSOLUTE_SUM": "The Position value or margin for each side (long and"
                        " short) of the Position are computed independently "
                        "and added together.",
        "MAXIMAL_SIDE": "The Position value or margin for each side (long and"
                        " short) of the Position are computed independently. "
                        "The Position value or margin chosen is the maximal "
                        "absolute value of the two.",
        "NET_SUM": "The units for each side (long and short) of the Position "
                   "are netted together and the resulting value (long or "
                   "short) is used to compute the Position value or margin."
    }
}



# -*- coding: utf-8 -*-
"""Handle account endpoints."""
from .apirequest import APIRequest
from .decorators import dyndoc_insert, endpoint
from .responses.accounts import responses
from abc import abstractmethod


class Accounts(APIRequest):
    """Accounts - class to handle the accounts endpoints."""

    ENDPOINT = ""
    METHOD = "GET"

    @abstractmethod
    @dyndoc_insert(responses)
    def __init__(self, accountID=None):
        """Instantiate an Accounts APIRequest instance.

        Parameters
        ----------
        accountID : string (optional)
            the accountID of the account. Optional when requesting
            all accounts. For all other requests to the endpoint it is
            required.

        """
        endpoint = self.ENDPOINT.format(accountID=accountID)
        super(Accounts, self).__init__(endpoint, method=self.METHOD)


@endpoint("v3/accounts")
class AccountList(Accounts):
    """Get a list of all Accounts authorized for the provided token."""

    @dyndoc_insert(responses)
    def __init__(self):
        """Instantiate an AccountList request.

        >>> import oandapyV20
        >>> import oandapyV20.endpoints.accounts as accounts
        >>> client = oandapyV20.API(access_token=...)
        >>> r = accounts.AccountList()
        >>> client.request(r)
        >>> print r.response

        ::

            {_v3_accounts_resp}

        """
        super(AccountList, self).__init__()


@endpoint("v3/accounts/{accountID}")
class AccountDetails(Accounts):
    """AccountDetails.

    Get the full details for a single Account that a client has access
    to. Full pending Order, open Trade and open Position representations are
    provided.
    """

    @dyndoc_insert(responses)
    def __init__(self, accountID):
        """Instantiate an AccountDetails request.

        Parameters
        ----------
        accountID : string (required)
            id of the account to perform the request on.


        >>> import oandapyV20
        >>> import oandapyV20.endpoints.accounts as accounts
        >>> client = oandapyV20.API(access_token=...)
        >>> r = accounts.AccountDetails(accountID)
        >>> client.request(r)
        >>> print r.response

        ::

            {_v3_account_by_accountID_resp}

        """
        super(AccountDetails, self).__init__(accountID)


@endpoint("v3/accounts/{accountID}/summary")
class AccountSummary(Accounts):
    """Get a summary for a single Account that a client has access to."""

    @dyndoc_insert(responses)
    def __init__(self, accountID):
        """Instantiate an AccountSummary request.

        Parameters
        ----------
        accountID : string (required)
            id of the account to perform the request on.


        >>> import oandapyV20
        >>> import oandapyV20.endpoints.accounts as accounts
        >>> client = oandapyV20.API(access_token=...)
        >>> r = accounts.AccountSummary(accountID)
        >>> client.request(r)
        >>> print r.response

        ::

            {_v3_account_by_accountID_summary_resp}

        """
        super(AccountSummary, self).__init__(accountID)


@endpoint("v3/accounts/{accountID}/instruments")
class AccountInstruments(Accounts):
    """AccountInstruments.

    Get the list of tradable instruments for the given Account. The list of
    tradeable instruments is dependent on the regulatory division that the
    Account is located in, thus should be the same for all Accounts owned by a
    single user.
    """

    @dyndoc_insert(responses)
    def __init__(self, accountID, params=None):
        """Instantiate an AccountInstruments request.

        Parameters
        ----------
        accountID : string (required)
            id of the account to perform the request on.

        params : dict (optional)
            query params to send, check developer.oanda.com for details.


        Query Params example::

            {_v3_account_by_accountID_instruments_params}


        >>> import oandapyV20
        >>> import oandapyV20.endpoints.accounts as accounts
        >>> client = oandapyV20.API(access_token=...)
        >>> params = ...
        >>> r = accounts.AccountInstruments(accountID=..., params=params)
        >>> client.request(r)
        >>> print r.response

        Output::

            {_v3_account_by_accountID_instruments_resp}

        """
        super(AccountInstruments, self).__init__(accountID)
        self.params = params


@endpoint("v3/accounts/{accountID}/configuration", "PATCH")
class AccountConfiguration(Accounts):
    """Set the client-configurable portions of an Account."""

    HEADERS = {"Content-Type": "application/json"}

    @dyndoc_insert(responses)
    def __init__(self, accountID, data):
        """Instantiate an AccountConfiguration request.

        Parameters
        ----------
        accountID : string (required)
            id of the account to perform the request on.

        data : dict (required)
            json body to send


        body example::

            {_v3_accounts_accountID_account_config_body}

        >>> import oandapyV20
        >>> import oandapyV20.endpoints.accounts as accounts
        >>> client = oandapyV20.API(access_token=...)
        >>> r = accounts.AccountConfiguration(accountID, data=data)
        >>> client.request(r)
        >>> print r.response

        ::

            {_v3_accounts_accountID_account_config_resp}

        """
        super(AccountConfiguration, self).__init__(accountID)
        self.data = data


@endpoint("v3/accounts/{accountID}/changes")
class AccountChanges(Accounts):
    """AccountChanges.

    Endpoint used to poll an Account for its current state and changes
    since a specified TransactionID.
    """

    @dyndoc_insert(responses)
    def __init__(self, accountID, params=None):
        """Instantiate an AccountChanges request.

        Parameters
        ----------
        accountID : string (required)
            id of the account to perform the request on.

        params : dict (optional)
            query params to send, check developer.oanda.com for details.


        Query Params example::

            {_v3_accounts_accountID_account_changes_params}

        >>> import oandapyV20
        >>> import oandapyV20.endpoints.accounts as accounts
        >>> client = oandapyV20.API(access_token=...)
        >>> params = ...
        >>> r = accounts.AccountChanges(accountID=..., params=params)
        >>> client.request(r)
        >>> print r.response

        Output::

            {_v3_accounts_accountID_account_changes_resp}

        """
        super(AccountChanges, self).__init__(accountID)
        self.params = params



# -*- coding: utf-8 -*-
"""Pricing related Definitions."""

definitions = {
    "PriceStatus": {
        "tradeable": "The Instrument’s price is tradeable.",
        "non-tradeable": "The Instrument’s price is not tradeable.",
        "invalid": "The Instrument of the price is invalid or there "
                   "is no valid Price for the Instrument."
    },
}



"""Responses.

responses serve both testing purpose aswell as dynamic docstring replacement
"""
responses = {
    "_v3_accounts": {
        "url": "/v3/accounts",
        "response": {
            "accounts": [
                {
                  "id": "101-004-1435156-002",
                  "tags": []
                },
                {
                  "id": "101-004-1435156-001",
                  "tags": []
                }
            ]
            },
        },
    "_v3_account_by_accountID": {
        "url": "/v3/accounts/{}",
        "response": {
            "account": {
                "trades": [
                    {
                        "instrument": "DE30_EUR",
                        "financing": "0.0000",
                        "openTime": "2016-07-12T09:32:18.062823776Z",
                        "initialUnits": "-10",
                        "currentUnits": "-10",
                        "price": "9984.7",
                        "unrealizedPL": "341.0000",
                        "realizedPL": "0.0000",
                        "state": "OPEN",
                        "id": "821"
                    },
                    {
                        "instrument": "DE30_EUR",
                        "financing": "0.0000",
                        "openTime": "2016-07-12T09:32:18.206929733Z",
                        "initialUnits": "-10",
                        "currentUnits": "-10",
                        "price": "9984.7",
                        "unrealizedPL": "341.0000",
                        "realizedPL": "0.0000",
                        "state": "OPEN",
                        "id": "823"
                    }
                ],
                "marginCloseoutNAV": "49393.6580",
                "marginUsed": "9948.9000",
                "currency": "EUR",
                "resettablePL": "-1301.0046",
                "NAV": "49377.6580",
                "marginCloseoutMarginUsed": "9949.8000",
                "id": "101-004-1435156-001",
                "marginCloseoutPositionValue": "198996.0000",
                "openTradeCount": 2,
                "orders": [
                    {
                        "partialFill": "DEFAULT_FILL",
                        "price": "0.87000",
                        "stopLossOnFill": {
                            "timeInForce": "GTC",
                            "price": "0.88000"
                        },
                        "timeInForce": "GTC",
                        "clientExtensions": {
                            "comment": "myComment",
                            "id": "myID"
                        },
                        "id": "204",
                        "triggerCondition": "TRIGGER_DEFAULT",
                        "replacesOrderID": "200",
                        "positionFill": "POSITION_DEFAULT",
                        "createTime": "2016-07-08T07:18:47.623211321Z",
                        "instrument": "EUR_GBP",
                        "state": "PENDING",
                        "units": "-50000",
                        "type": "LIMIT"
                    }
                ],
                "hedgingEnabled": False,
                "marginCloseoutPercent": "0.10072",
                "marginCallMarginUsed": "9949.8000",
                "openPositionCount": 1,
                "positionValue": "198978.0000",
                "pl": "-1301.0046",
                "lastTransactionID": "833",
                "marginAvailable": "39428.7580",
                "marginCloseoutUnrealizedPL": "698.0000",
                "marginRate": "0.05",
                "marginCallPercent": "0.20144",
                "pendingOrderCount": 1,
                "withdrawalLimit": "39428.7580",
                "unrealizedPL": "682.0000",
                "alias": "hootnotv20",
                "createdByUserID": 1435156,
                "positions": [
                    {
                        "short": {
                            "units": "0",
                            "resettablePL": "0.0000",
                            "unrealizedPL": "0.0000",
                            "pl": "0.0000"
                        },
                        "unrealizedPL": "0.0000",
                        "long": {
                            "units": "0",
                            "resettablePL": "-3.8046",
                            "unrealizedPL": "0.0000",
                            "pl": "-3.8046"
                        },
                        "instrument": "EUR_USD",
                        "resettablePL": "-3.8046",
                        "pl": "-3.8046"
                    },
                    {
                        "short": {
                            "unrealizedPL": "682.0000",
                            "tradeIDs": [
                                "821",
                                "823"
                            ],
                            "resettablePL": "-1744.8000",
                            "units": "-20",
                            "averagePrice": "9984.7",
                            "pl": "-1744.8000"
                        },
                        "unrealizedPL": "682.0000",
                        "long": {
                            "units": "0",
                            "resettablePL": "447.6000",
                            "unrealizedPL": "0.0000",
                            "pl": "447.6000"
                        },
                        "instrument": "DE30_EUR",
                        "resettablePL": "-1297.2000",
                        "pl": "-1297.2000"
                    }
                ],
                "createdTime": "2016-06-24T21:03:50.914647476Z",
                "balance": "48695.6580"
            },
            "lastTransactionID": "833"
        }
    },
    "_v3_account_by_accountID_summary": {
        "url": "v3/accounts/{accountID}/summary",
        "response": {
            "account": {
                "marginCloseoutNAV": "35454.4740",
                "marginUsed": "10581.5000",
                "currency": "EUR",
                "resettablePL": "-13840.3525",
                "NAV": "35454.4740",
                "marginCloseoutMarginUsed": "10581.5000",
                "marginCloseoutPositionValue": "211630.0000",
                "openTradeCount": 2,
                "id": "101-004-1435156-001",
                "openPositionCount": 1,
                "marginCloseoutPercent": "0.14923",
                "marginCallMarginUsed": "10581.5000",
                "hedgingEnabled": False,
                "positionValue": "211630.0000",
                "pl": "-13840.3525",
                "lastTransactionID": "2123",
                "marginAvailable": "24872.9740",
                "marginRate": "0.05",
                "marginCallPercent": "0.29845",
                "pendingOrderCount": 0,
                "withdrawalLimit": "24872.9740",
                "unrealizedPL": "0.0000",
                "alias": "hootnotv20",
                "createdByUserID": 1435156,
                "marginCloseoutUnrealizedPL": "0.0000",
                "createdTime": "2016-06-24T21:03:50.914647476Z",
                "balance": "35454.4740"
            },
            "lastTransactionID": "2123"
        }
    },
    "_v3_account_by_accountID_instruments": {
        "url": "/v3/accounts/{accountID}/instuments",
        "params": {
            "instruments": "EU50_EUR,EUR_USD,US30_USD,"
                           "FR40_EUR,EUR_CHF,DE30_EUR"
        },
        "response": {
            "instruments": [
                {
                    "minimumTradeSize": "1",
                    "displayName": "Europe 50",
                    "name": "EU50_EUR",
                    "displayPrecision": 1,
                    "type": "CFD",
                    "minimumTrailingStopDistance": "5.0",
                    "marginRate": "0.05",
                    "maximumOrderUnits": "3000",
                    "tradeUnitsPrecision": 0,
                    "pipLocation": 0,
                    "maximumPositionSize": "0",
                    "maximumTrailingStopDistance": "10000.0"
                },
                {
                    "minimumTradeSize": "1",
                    "displayName": "EUR/USD",
                    "name": "EUR_USD",
                    "displayPrecision": 5,
                    "type": "CURRENCY",
                    "minimumTrailingStopDistance": "0.00050",
                    "marginRate": "0.05",
                    "maximumOrderUnits": "100000000",
                    "tradeUnitsPrecision": 0,
                    "pipLocation": -4,
                    "maximumPositionSize": "0",
                    "maximumTrailingStopDistance": "1.00000"
                },
                {
                    "minimumTradeSize": "1",
                    "displayName": "US Wall St 30",
                    "name": "US30_USD",
                    "displayPrecision": 1,
                    "type": "CFD",
                    "minimumTrailingStopDistance": "5.0",
                    "marginRate": "0.05",
                    "maximumOrderUnits": "1000",
                    "tradeUnitsPrecision": 0,
                    "pipLocation": 0,
                    "maximumPositionSize": "0",
                    "maximumTrailingStopDistance": "10000.0"
                },
                {
                    "minimumTradeSize": "1",
                    "displayName": "France 40",
                    "name": "FR40_EUR",
                    "displayPrecision": 1,
                    "type": "CFD",
                    "minimumTrailingStopDistance": "5.0",
                    "marginRate": "0.05",
                    "maximumOrderUnits": "2000",
                    "tradeUnitsPrecision": 0,
                    "pipLocation": 0,
                    "maximumPositionSize": "0",
                    "maximumTrailingStopDistance": "10000.0"
                },
                {
                    "minimumTradeSize": "1",
                    "displayName": "EUR/CHF",
                    "name": "EUR_CHF",
                    "displayPrecision": 5,
                    "type": "CURRENCY",
                    "minimumTrailingStopDistance": "0.00050",
                    "marginRate": "0.05",
                    "maximumOrderUnits": "100000000",
                    "tradeUnitsPrecision": 0,
                    "pipLocation": -4,
                    "maximumPositionSize": "0",
                    "maximumTrailingStopDistance": "1.00000"
                },
                {
                    "minimumTradeSize": "1",
                    "displayName": "Germany 30",
                    "name": "DE30_EUR",
                    "displayPrecision": 1,
                    "type": "CFD",
                    "minimumTrailingStopDistance": "5.0",
                    "marginRate": "0.05",
                    "maximumOrderUnits": "2500",
                    "tradeUnitsPrecision": 0,
                    "pipLocation": 0,
                    "maximumPositionSize": "0",
                    "maximumTrailingStopDistance": "10000.0"
                },
            ],
            "lastTransactionID": "2124"
        },
    },
    "_v3_accounts_accountID_account_config": {
        "url": "/v3/accounts/{accountID}/configuration",
        "body": {
            "marginRate": "0.05"
        },
        "response": {
            "lastTransactionID": "830",
            "clientConfigureTransaction": {
                "userID": 1435156,
                "marginRate": "0.05",
                "batchID": "830",
                "time": "2016-07-12T19:48:11.657494168Z",
                "type": "CLIENT_CONFIGURE",
                "id": "830",
                "accountID": "101-004-1435156-001"
            }
        },
    },
    "_v3_accounts_accountID_account_changes": {
        "url": "/v3/accounts/{accountID}/changes",
        "params": {
            "sinceTransactionID": 2308
        },
        "response": {
            "state": {
                "trades": [],
                "marginCloseoutNAV": "33848.2663",
                "unrealizedPL": "0.0000",
                "marginUsed": "0.0000",
                "marginAvailable": "33848.2663",
                "positions": [],
                "marginCloseoutUnrealizedPL": "0.0000",
                "marginCallMarginUsed": "0.0000",
                "marginCallPercent": "0.00000",
                "marginCloseoutPercent": "0.00000",
                "NAV": "33848.2663",
                "marginCloseoutMarginUsed": "0.0000",
                "positionValue": "0.0000",
                "orders": [],
                "withdrawalLimit": "33848.2663"
            },
            "changes": {
                "tradesReduced": [],
                "tradesOpened": [],
                "ordersFilled": [],
                "transactions": [
                  {
                    "timeInForce": "GTC",
                    "triggerCondition": "TRIGGER_DEFAULT",
                    "positionFill": "DEFAULT",
                    "stopLossOnFill": {
                      "timeInForce": "GTC",
                      "price": "1.22000"
                    },
                    "userID": 1435156,
                    "id": "2309",
                    "batchID": "2309",
                    "instrument": "EUR_USD",
                    "reason": "CLIENT_ORDER",
                    "time": "2016-10-25T21:07:21.065554321Z",
                    "units": "-100",
                    "type": "LIMIT_ORDER",
                    "price": "1.20000",
                    "accountID": "101-004-1435156-001"
                  }
                ],
                "ordersCreated": [
                  {
                    "triggerCondition": "TRIGGER_DEFAULT",
                    "partialFill": "DEFAULT_FILL",
                    "price": "1.20000",
                    "stopLossOnFill": {
                      "timeInForce": "GTC",
                      "price": "1.22000"
                    },
                    "createTime": "2016-10-25T21:07:21.065554321Z",
                    "timeInForce": "GTC",
                    "instrument": "EUR_USD",
                    "state": "PENDING",
                    "units": "-100",
                    "id": "2309",
                    "type": "LIMIT",
                    "positionFill": "POSITION_DEFAULT"
                  }
                ],
                "positions": [],
                "ordersTriggered": [],
                "ordersCancelled": [],
                "tradesClosed": []
            },
            "lastTransactionID": "2309"
        }
    }
}



# -*- coding: utf-8 -*-
"""Responses.

responses serve both testing purpose aswell as dynamic docstring replacement
"""
responses = {
    "_v3_accounts_accountID_orders_create": {
        "url": "v3/accounts/{accountID}/orders",
        "body": {
                "order": {
                    "stopLossOnFill": {
                        "timeInForce": "GTC",
                        "price": "1.22"
                    },
                    "units": "-100",
                    "price": "1.2",
                    "instrument": "EUR_USD",
                    "timeInForce": "GTC",
                    "type": "LIMIT",
                    "positionFill": "DEFAULT"
                }
        },
        "response": {
            "orderCreateTransaction": {
                "timeInForce": "GTC",
                "triggerCondition": "TRIGGER_DEFAULT",
                "positionFill": "DEFAULT",
                "stopLossOnFill": {
                    "timeInForce": "GTC",
                    "price": "1.22000"
                },
                "userID": 1435156,
                "id": "2304",
                "batchID": "2304",
                "instrument": "EUR_USD",
                "reason": "CLIENT_ORDER",
                "time": "2016-10-24T21:48:18.593753865Z",
                "units": "-100",
                "type": "LIMIT_ORDER",
                "price": "1.20000",
                "accountID": "101-004-1435156-001"
            },
            "relatedTransactionIDs": [
                "2304"
            ],
            "lastTransactionID": "2304"
        }
    },
    "_v3_accounts_accountID_orders_pending": {
        "url": "v3/accounts/{accountID}/orders",
        "response": {
            "orders": [
                {
                    "timeInForce": "GTC",
                    "triggerCondition": "TRIGGER_DEFAULT",
                    "partialFill": "DEFAULT_FILL",
                    "positionFill": "POSITION_DEFAULT",
                    "stopLossOnFill": {
                        "timeInForce": "GTC",
                        "price": "1.22000"
                    },
                    "id": "2304",
                    "price": "1.20000",
                    "instrument": "EUR_USD",
                    "state": "PENDING",
                    "units": "-100",
                    "clientExtensions": {
                        "comment": "myComment",
                        "id": "myID"
                    },
                    "type": "LIMIT",
                    "createTime": "2016-10-24T21:48:18.593753865Z"
                }
            ],
            "lastTransactionID": "2305"
        }
    },
    "_v3_accounts_accountID_order_cancel": {
        "url": "v3/accounts/{accountID}/orders/{orderID}/cancel",
        "orderID": "2307",
        "response": {
            "orderCancelTransaction": {
                "orderID": "2307",
                "clientOrderID": "myID",
                "userID": 1435156,
                "batchID": "2308",
                "reason": "CLIENT_REQUEST",
                "time": "2016-10-25T20:53:03.789670387Z",
                "type": "ORDER_CANCEL",
                "id": "2308",
                "accountID": "101-004-1435156-001"
            },
            "relatedTransactionIDs": [
                "2308"
            ],
            "lastTransactionID": "2308"
        }
    },
    "_v3_accounts_accountID_orders_list": {
        "url": "v3/accounts/{accountID}/orders",
        "response": {
            "orders": [
                {
                    "triggerCondition": "TRIGGER_DEFAULT",
                    "partialFill": "DEFAULT_FILL",
                    "price": "1.20000",
                    "stopLossOnFill": {
                        "timeInForce": "GTC",
                        "price": "1.22000"
                    },
                    "createTime": "2016-10-05T10:25:47.627003645Z",
                    "timeInForce": "GTC",
                    "instrument": "EUR_USD",
                    "state": "PENDING",
                    "units": "-100",
                    "id": "2125",
                    "type": "LIMIT",
                    "positionFill": "POSITION_DEFAULT"
                }
            ],
            "lastTransactionID": "2129"
        }
    },
    "_v3_accounts_accountID_order_details": {
        "url": "v3/accounts/{accountID}/orders/{orderID}/details",
        "orderID": "2309",
        "response": {
            "order": {
                "triggerCondition": "TRIGGER_DEFAULT",
                "partialFill": "DEFAULT_FILL",
                "price": "1.20000",
                "stopLossOnFill": {
                    "timeInForce": "GTC",
                    "price": "1.22000"
                },
                "createTime": "2016-10-25T21:07:21.065554321Z",
                "timeInForce": "GTC",
                "instrument": "EUR_USD",
                "state": "PENDING",
                "units": "-100",
                "id": "2309",
                "type": "LIMIT",
                "positionFill": "POSITION_DEFAULT"
            },
            "lastTransactionID": "2309"
        }
    },
    "_v3_accounts_accountID_order_replace": {
        "url": "v3/accounts/{accountID}/orders",
        "body": {
            "order": {
                "units": "-500000",
                "type": "LIMIT",
                "instrument": "EUR_USD",
                "price": "1.25000",
            }
        },
        "response": {
                "orderCreateTransaction": {
                    "timeInForce": "GTC",
                    "triggerCondition": "TRIGGER_DEFAULT",
                    "replacesOrderID": "2304",
                    "positionFill": "DEFAULT",
                    "userID": 1435156,
                    "units": "-500000",
                    "batchID": "2306",
                    "instrument": "EUR_USD",
                    "reason": "REPLACEMENT",
                    "time": "2016-10-25T19:45:38.558056359Z",
                    "price": "1.25000",
                    "clientExtensions": {
                        "comment": "myComment",
                        "id": "myID"
                    },
                    "type": "LIMIT_ORDER",
                    "id": "2307",
                    "accountID": "101-004-1435156-001"
                },
                "orderCancelTransaction": {
                    "orderID": "2304",
                    "clientOrderID": "myID",
                    "replacedByOrderID": "2307",
                    "userID": 1435156,
                    "batchID": "2306",
                    "reason": "CLIENT_REQUEST_REPLACED",
                    "time": "2016-10-25T19:45:38.558056359Z",
                    "type": "ORDER_CANCEL",
                    "id": "2306",
                    "accountID": "101-004-1435156-001"
                },
                "relatedTransactionIDs": [
                    "2306",
                    "2307"
                ],
                "lastTransactionID": "2307"
        }
    },
    "_v3_accounts_accountID_order_clientextensions": {
        "url": "v3/accounts/{accountID}/orders/{orderID}/clientExtensions",
        "orderID": 2304,
        "body": {
            "clientExtensions": {
                "id": "myID",
                "comment": "myComment",
            }
        },
        "response": {
            "relatedTransactionIDs": [
                "2305"
            ],
            "orderClientExtensionsModifyTransaction": {
                "orderID": "2304",
                "userID": 1435156,
                "batchID": "2305",
                "clientExtensionsModify": {
                    "comment": "myComment",
                    "id": "myID"
                },
                "time": "2016-10-25T15:56:43.075594239Z",
                "type": "ORDER_CLIENT_EXTENSIONS_MODIFY",
                "id": "2305",
                "accountID": "101-004-1435156-001"
            },
            "lastTransactionID": "2305"
        }
    }
}



# -*- coding: utf-8 -*-
"""OANDA API wrapper for OANDA's REST-V20 API."""

import json
import requests
import logging
from .exceptions import V20Error

ITER_LINES_CHUNKSIZE = 60

TRADING_ENVIRONMENTS = {
    "practice": {
        "stream": 'https://stream-fxpractice.oanda.com',
        "api": 'https://api-fxpractice.oanda.com'
    },
    "live": {
        "stream": 'https://stream-fxtrade.oanda.com',
        "api": 'https://api-fxtrade.oanda.com'
    }
}

DEFAULT_HEADERS = {
    "Accept-Encoding": "gzip, deflate"
}

logger = logging.getLogger(__name__)


class API(object):
    r"""API - class to handle APIRequests objects to access API endpoints.

    Examples
    --------

    ::

        # get a list of trades
        from oandapyV20 import API
        import oandapyV20.endpoints.trades as trades

        api = API(access_token="xxx")
        accountID = "101-305-3091856-001"

        r = trades.TradesList(accountID)
        # show the endpoint as it is constructed for this call
        print("REQUEST:{}".format(r))
        rv = api.request(r)
        print("RESPONSE:\n{}".format(json.dumps(rv, indent=2)))


    Output::

        REQUEST:v3/accounts/101-305-3091856-001/trades
        RESPONSE:
        "trades": [
            {
              "financing": "0.0000",
              "openTime": "2016-07-21T15:47:05.170212014Z",
              "price": "10133.9",
              "unrealizedPL": "8.0000",
              "realizedPL": "0.0000",
              "instrument": "DE30_EUR",
              "state": "OPEN",
              "initialUnits": "-10",
              "currentUnits": "-10",
              "id": "1032"
            },
            {
              "financing": "0.0000",
              "openTime": "2016-07-21T15:47:04.963590941Z",
              "price": "10134.4",
              "unrealizedPL": "13.0000",
              "realizedPL": "0.0000",
              "instrument": "DE30_EUR",
              "state": "OPEN",
              "initialUnits": "-10",
              "currentUnits": "-10",
              "id": "1030"
            }
          ],
          "lastTransactionID": "1040"
        }

    ::

        # reduce a trade by it's id
        from oandapyV20 import API
        import oandapyV20.endpoints.trades as trades

        api = API(access_token="...")

        accountID = "101-305-3091856-001"
        tradeID = "1030"
        cfg = {"units": 5}
        r = trades.TradeClose(accountID, tradeID=tradeID, data=cfg)
        # show the endpoint as it is constructed for this call
        print("REQUEST:{}".format(r))
        rv = api.request(r)
        print("RESPONSE\n{}".format(json.dumps(rv, indent=2)))


    Output::

        REQUEST:v3/accounts/101-305-3091856-001/trades/1030/close
        RESPONSE: {
          "orderFillTransaction": {
            "orderID": "1041",
            "financing": "-0.1519",
            "instrument": "DE30_EUR",
            "userID": 1435156,
            "price": "10131.6",
            "tradeReduced": {
              "units": "5",
              "financing": "-0.1519",
              "realizedPL": "14.0000",
              "tradeID": "1030"
            },
            "batchID": "1041",
            "accountBalance": "44876.2548",
            "reason": "MARKET_ORDER_TRADE_CLOSE",
            "time": "2016-07-21T17:32:51.361464739Z",
            "units": "5",
            "type": "ORDER_FILL",
            "id": "1042",
            "pl": "14.0000",
            "accountID": "101-305-3091856-001"
          },
          "orderCreateTransaction": {
            "timeInForce": "FOK",
            "positionFill": "REDUCE_ONLY",
            "userID": 1435156,
            "batchID": "1041",
            "instrument": "DE30_EUR",
            "reason": "TRADE_CLOSE",
            "tradeClose": {
              "units": "5",
              "tradeID": "1030"
            },
            "time": "2016-07-21T17:32:51.361464739Z",
            "units": "5",
            "type": "MARKET_ORDER",
            "id": "1041",
            "accountID": "101-305-3091856-001"
          },
          "relatedTransactionIDs": [
            "1041",
            "1042"
          ],
          "lastTransactionID": "1042"
        }
    """

    def __init__(self, access_token, environment="practice",
                 headers=None, request_params=None):
        """Instantiate an instance of OandaPy's API wrapper.

        Parameters
        ----------
        access_token : string
            Provide a valid access token.

        environment : string
            Provide the environment for OANDA's REST api. Valid values:
            'practice' or 'live'. Default: 'practice'.

        headers : dict (optional)
            Provide request headers to be set for a request.


        .. note::

            There is no need to set the 'Content-Type: application/json'
            for the endpoints that require this header. The API-request
            classes covering those endpoints will take care of the header.

        request_params : (optional)
            parameters to be passed to the request. This can be used to apply
            for instance a timeout value:

               request_params={"timeout": 0.1}

            See specs of the requests module for full details of possible
            parameters.

        .. warning::
            parameters belonging to a request need to be set on the
            requestinstance and are NOT passed via the client.

        """
        logger.info("setting up API-client for environment %s", environment)
        try:
            TRADING_ENVIRONMENTS[environment]
        except:
            logger.error("unkown environment %s", environment)
            raise KeyError("Unknown environment: {}".format(environment))
        else:
            self.environment = environment

        self.access_token = access_token
        self.client = requests.Session()
        self.client.stream = False
        self._request_params = request_params if request_params else {}

        # personal token authentication
        if self.access_token:
            self.client.headers['Authorization'] = 'Bearer '+self.access_token

        self.client.headers.update(DEFAULT_HEADERS)
        if headers:
            self.client.headers.update(headers)
            logger.info("applying headers %s", ",".join(headers.keys()))

    @property
    def request_params(self):
        """request_params property."""
        return self._request_params

    def __request(self, method, url, request_args, headers=None, stream=False):
        """__request.

        make the actual request. This method is called by the
        request method in case of 'regular' API-calls. Or indirectly by
        the__stream_request method if it concerns a 'streaming' call.
        """
        func = getattr(self.client, method)
        headers = headers if headers else {}
        response = None
        try:
            logger.info("performing request %s", url)
            response = func(url, stream=stream, headers=headers,
                            **request_args)
        except requests.RequestException as err:
            logger.error("request %s failed [%s]", url, err)
            raise err

        # Handle error responses
        if response.status_code >= 400:
            logger.error("request %s failed [%d,%s]",
                         url,
                         response.status_code,
                         response.content.decode('utf-8'))
            raise V20Error(response.status_code,
                           response.content.decode('utf-8'))
        return response

    def __stream_request(self, method, url, request_args, headers=None):
        """__stream_request.

        make a 'stream' request. This method is called by
        the 'request' method after it has determined which
        call applies: regular or streaming.
        """
        headers = headers if headers else {}
        response = self.__request(method, url, request_args,
                                  headers=headers, stream=True)
        lines = response.iter_lines(ITER_LINES_CHUNKSIZE)
        for line in lines:
            if line:
                data = json.loads(line.decode("utf-8"))
                yield data

    def request(self, endpoint):
        """Perform a request for the APIRequest instance 'endpoint'.

        Parameters
        ----------
        endpoint : APIRequest
            The endpoint parameter contains an instance of an APIRequest
            containing the endpoint, method and optionally other parameters
            or body data.

        Raises
        ------
            V20Error in case of HTTP response code >= 400
        """
        method = endpoint.method
        method = method.lower()
        params = None
        try:
            params = getattr(endpoint, "params")
        except AttributeError:
            # request does not have params
            params = {}

        headers = {}
        if hasattr(endpoint, "HEADERS"):
            headers = getattr(endpoint, "HEADERS")

        request_args = {}
        if method == 'get':
            request_args['params'] = params
        elif hasattr(endpoint, "data") and endpoint.data:
            request_args['json'] = endpoint.data

        # if any parameter for request then merge them
        request_args.update(self._request_params)

        # which API to access ?
        if not (hasattr(endpoint, "STREAM") and
                getattr(endpoint, "STREAM") is True):
            url = "{}/{}".format(
                TRADING_ENVIRONMENTS[self.environment]["api"],
                endpoint)

            response = self.__request(method, url,
                                      request_args, headers=headers)
            content = response.content.decode('utf-8')
            content = json.loads(content)

            # update endpoint
            endpoint.response = content
            endpoint.status_code = response.status_code

            return content

        else:
            url = "{}/{}".format(
                TRADING_ENVIRONMENTS[self.environment]["stream"],
                endpoint)
            endpoint.response = self.__stream_request(method,
                                                      url,
                                                      request_args,
                                                      headers=headers)
            return endpoint.response



import logging
from .oandapyV20 import API
from .exceptions import V20Error

__title__ = "OANDA REST V20 API Wrapper"
__version__ = "0.6.1"
__author__ = "Feite Brekeveld"
__license__ = "MIT"
__copyright__ = "Copyright 2016 - 2018 Feite Brekeveld"

# Version synonym
VERSION = __version__

# Set default logging handler to avoid "No handler found" warnings.
try:
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

logging.getLogger(__name__).addHandler(NullHandler())

__all__ = (
    'API',
    'V20Error'
)



# -*- coding: utf-8 -*-

from datetime import datetime
import calendar
import logging

import oandapyV20.endpoints.instruments as instruments
from oandapyV20.contrib.generic import granularity_to_time, secs2time


logger = logging.getLogger(__name__)

MAX_BATCH = 5000
DEFAULT_BATCH = 500


def InstrumentsCandlesFactory(instrument, params=None):
    """InstrumentsCandlesFactory - generate InstrumentCandles requests.

    InstrumentsCandlesFactory is used to retrieve historical data by
    automatically generating consecutive requests when the OANDA limit
    of *count* records is exceeded.

    This is known by calculating the number of candles between *from* and
    *to*. If *to* is not specified *to* will be equal to *now*.

    The *count* parameter is only used to control the number of records to
    retrieve in a single request.

    The *includeFirst* parameter is forced to make sure that results do
    no have a 1-record gap between consecutive requests.

    Parameters
    ----------

    instrument : string (required)
        the instrument to create the order for

    params: params (optional)
        the parameters to specify the historical range,
        see the REST-V20 docs regarding 'instrument' at developer.oanda.com
        If no params are specified, just a single InstrumentsCandles request
        will be generated acting the same as if you had just created it
        directly.

    Example
    -------

    The *oandapyV20.API* client processes requests as objects. So,
    downloading large historical batches simply comes down to:

    >>> import json
    >>> from oandapyV20 import API
    >>> from oandapyV20.contrib.factories import InstrumentsCandlesFactory
    >>>
    >>> client = API(access_token=...)
    >>> instrument, granularity = "EUR_USD", "M15"
    >>> _from = "2017-01-01T00:00:00Z"
    >>> params = {
    ...    "from": _from,
    ...    "granularity": granularity,
    ...    "count": 2500,
    ... }
    >>> with open("/tmp/{}.{}".format(instrument, granularity), "w") as OUT:
    >>>     # The factory returns a generator generating consecutive
    >>>     # requests to retrieve full history from date 'from' till 'to'
    >>>     for r in InstrumentsCandlesFactory(instrument=instrument,
    ...                                        params=params)
    >>>         client.request(r)
    >>>         OUT.write(json.dumps(r.response.get('candles'), indent=2))

    .. note:: Normally you can't combine *from*, *to* and *count*.
              When *count* specified, it is used to calculate the gap between
              *to* and *from*. The *params* passed to the generated request
              itself does contain the *count* parameter.

    """
    RFC3339 = "%Y-%m-%dT%H:%M:%SZ"

    # if not specified use the default of 'S5' as OANDA does
    gs = granularity_to_time(params.get('granularity', 'S5'))

    _from = None
    _epoch_from = None
    if 'from' in params:
        _from = datetime.strptime(params.get('from'), RFC3339)
        _epoch_from = int(calendar.timegm(_from.timetuple()))

    _to = datetime.utcnow()
    if 'to' in params:
        _tmp = datetime.strptime(params.get('to'), RFC3339)
        # if specified datetime > now, we use 'now' instead
        if _tmp > _to:
            logger.info("datetime %s is in the future, will be set to 'now'",
                        params.get('to'))
        else:
            _to = _tmp

    _epoch_to = int(calendar.timegm(_to.timetuple()))

    _count = params.get('count', DEFAULT_BATCH)
    # OANDA will respond with a V20Error if count > MAX_BATCH

    if 'to' in params and 'from' not in params:
        raise ValueError("'to' specified without 'from'")

    if not params or 'from' not in params:
        yield instruments.InstrumentsCandles(instrument=instrument,
                                             params=params)

    else:
        delta = _epoch_to - _epoch_from
        nbars = delta / gs

        cpparams = params.copy()
        for k in ['count', 'from', 'to']:
            if k in cpparams:
                del cpparams[k]
        # force includeFirst
        cpparams.update({"includeFirst": True})

        # generate InstrumentsCandles requests for all 'bars', each request
        # requesting max. count records
        for _ in range(_count, int(((nbars//_count)+1))*_count+1, _count):
            to = _epoch_from + _count * gs
            if to > _epoch_to:
                to = _epoch_to
            yparams = cpparams.copy()
            yparams.update({"from": secs2time(_epoch_from).strftime(RFC3339)})
            yparams.update({"to": secs2time(to).strftime(RFC3339)})
            yield instruments.InstrumentsCandles(instrument=instrument,
                                                 params=yparams)
            _epoch_from = to



# -*- coding: utf-8 -*-
"""dynamically add the classes for definition representations.

Most of the endpoint groups have some definitions that apply. These are
in the definitions package. It is conveniant to have access by a class
representing a specific group of definitions instead of a dictionary.
"""
import sys
from importlib import import_module
import six


dyndoc = """Definition representation of {cls}

    Definitions used in requests and responses. This
    class provides the ID and the description of the definitions.

    >>> import {PTH} as def{mod}
    >>> print def{mod}.{cls}.{firstItem}
    {orig}
    >>> c = def{mod}.{cls}()
    >>> print c[c.{firstItem}]
    {firstItemVal}
    >>> # or
    >>> print def{mod}.{cls}().definitions[c.{firstItem}]
    >>> # all keys
    >>> print def{mod}.{cls}().definitions.keys()
    >>> ...

"""

_doc = """
    .. note::

       attribute name *{}* is renamed to *{}*, value stil is *{}*. This
       means that a lookup stil applies.
"""


def make_definition_classes(mod):
    """Dynamically create the definition classes from module 'mod'."""
    rootpath = "oandapyV20"
    PTH = "{}.definitions.{}".format(rootpath, mod)

    M = import_module(PTH)
    __ALL__ = []  # construct the __all__ variable
    for cls, cldef in M.definitions.items():

        orig, fiV = next(six.iteritems(cldef))
        fiK = orig.replace('-', '_')
        # create the docstring dynamically
        clsdoc = dyndoc.format(cls=cls,
                               PTH=PTH,
                               mod=mod,
                               firstItem=fiK, orig=orig,
                               firstItemVal=fiV)

        # Since we can't change the docstring afterwards (it's readonly)
        # figure this out before and not during ...
        for K, V in cldef.items():
            attrName = K
            if "-" in K:
                attrName = K.replace('-', '_')
                adoc = _doc.format(K, attrName, K)
                clsdoc += adoc

        # the class
        dyncls = type(cls, (object,), {'__doc__': clsdoc})

        definitions = dict()
        for K, V in cldef.items():
            attrName = K
            if "-" in K:
                attrName = K.replace('-', '_')
            setattr(dyncls, attrName, K)  # set as class attributes
            definitions.update({K: V})    # for mapping by __getitem__

        def mkgi():
            def __getitem__(self, definitionID):
                """return description for definitionID."""
                return self._definitions[definitionID]
            return __getitem__

        def mkinit(definitions):
            def __init__(self):
                self._definitions = definitions

            return __init__

        def mkPropDefinitions():
            def definitions(self):
                """readonly property holding definition dict."""
                return self._definitions
            return property(definitions)

        setattr(dyncls, "__getitem__", mkgi())
        setattr(dyncls, "__init__", mkinit(definitions))
        setattr(dyncls, "definitions", mkPropDefinitions())
        setattr(sys.modules["{}.definitions.{}".format(rootpath, mod)],
                cls, dyncls)
        __ALL__.append(cls)

    setattr(sys.modules["{}.definitions.{}".format(rootpath, mod)],
            "__all__", tuple(__ALL__))


definitionModules = [
    'accounts',
    'instruments',
    'orders',
    'pricing',
    'primitives',
    'trades',
    'transactions'
]

# dynamically create all the definition classes from the modules
for M in definitionModules:
    make_definition_classes(M)



# -*- coding: utf-8 -*-
"""positionclose."""

from .baserequest import BaseRequest
from oandapyV20.types import Units


class PositionCloseRequest(BaseRequest):
    """create a PositionCloseRequest.

    PositionCloseRequest is used to build the body to close a position.
    The body can be used to pass to the PositionClose endpoint.
    """

    def __init__(self,
                 longUnits=None,
                 longClientExtensions=None,
                 shortUnits=None,
                 shortClientExtensions=None):
        """
        Instantiate a PositionCloseRequest.

        Parameters
        ----------

        longUnits : integer (optional)
            the number of long units to close

        longClientExtensions : dict (optional)
            dict representing longClientExtensions

        shortUnits : integer (optional)
            the number of short units to close

        shortClientExtensions : dict (optional)
            dict representing shortClientExtensions


        One of the parameters or both must be supplied.

        Example
        -------

        >>> import json
        >>> from oandapyV20 import API
        >>> import oandapyV20.endpoints.positions as positions
        >>> from oandapyV20.contrib.requests import PositionCloseRequest
        >>>
        >>> accountID = "..."
        >>> client = API(access_token=...)
        >>> ordr = PositionCloseRequest(longUnits=10000)
        >>> print(json.dumps(ordr.data, indent=4))
        {
           "longUnits": "10000"
        }
        >>> # now we have the order specification, create the order request
        >>> r = position.PositionClose(accountID,
        >>>                            instrument="EUR_USD", data=ordr.data)
        >>> # perform the request
        >>> rv = client.request(r)
        >>> print(rv)
        >>> ...
        """
        super(PositionCloseRequest, self).__init__()

        if not (longUnits or shortUnits):
            raise ValueError("longUnits and/or shortUnits parameter required")

        if longUnits:
            self._data.update({"longUnits": Units(longUnits).value})

            if longClientExtensions:
                self._data.update({"longClientExtensions":
                                   longClientExtensions})

        if shortUnits:
            self._data.update({"shortUnits": Units(shortUnits).value})

            if shortClientExtensions:
                self._data.update({"shortClientExtensions":
                                   shortClientExtensions})



# -*- coding: utf-8 -*-
"""Trades definitions."""

definitions = {
    "TradeState": {
        "OPEN": "The Trade is currently open",
        "CLOSED": "The Trade has been fully closed",
        "CLOSE_WHEN_TRADABLE": "The Trade will be closed as soon as the "
                               "trade’s instrument becomes tradeable"
    },
    "TradeStateFilter": {
        "OPEN": "The Trades that are currently open",
        "CLOSED": "The Trades that have been fully closed",
        "CLOSE_WHEN_TRADEABLE": "The Trades that will be closed as soon as "
                                "the trades' instrument becomes tradeable",
        "ALL": "The Trades that are in any of the possible states listed "
               "above."
    },
    "TradePL": {
        "POSITIVE": "An open Trade currently has a positive (profitable) "
                    "unrealized P/L, or a closed Trade realized a positive "
                    "amount of P/L.",
        "NEGATIVE": "An open Trade currently has a negative (losing) "
                    "unrealized P/L, or a closed Trade realized a negative "
                    "amount of P/L",
        "ZERO": "An open Trade currently has unrealized P/L of zero "
                "(neither profitable nor losing), or a closed Trade realized "
                "a P/L amount of zero."
    }
}



"""Exceptions."""


class StreamTerminated(Exception):
    """StreamTerminated."""


class V20Error(Exception):
    """Generic error class.

    In case of HTTP response codes >= 400 this class can be used
    to raise an exception representing that error.
    """

    def __init__(self, code, msg):
        """Instantiate a V20Error.

        Parameters
        ----------
        code : int
            the HTTP-code of the response

        msg : str
            the message returned with the response
        """
        self.code = code
        self.msg = msg

        super(V20Error, self).__init__(msg)



# -*- coding: UTF-8 -*-
"""decorators."""


def dyndoc_insert(src):
    """docstring_insert - a decorator to insert API-docparts dynamically."""
    # manipulating docstrings this way is tricky due to indentation
    # the JSON needs leading whitespace to be interpreted correctly
    import json
    import re

    def mkblock(d, flag=0):
        # response, pretty formatted
        v = json.dumps(d, indent=2)
        if flag == 1:
            # strip the '[' and ']' in case of a list holding items
            # that stand on their own (example: tick records from a stream)
            nw = re.findall('.*?\[(.*)\]', v, flags=re.S)
            v = nw[0]
        # add leading whitespace for each line and start with a newline
        return "\n{}".format("".join(["{0:>16}{1}\n".format("", L)
                             for L in v.split('\n')]))

    def dec(obj):
        allSlots = re.findall("\{(_v3.*?)\}", obj.__doc__)
        docsub = {}
        sub = {}
        for k in allSlots:
            p = re.findall("^(_v3.*)_(.*)", k)
            p = list(*p)
            sub.update({p[1]: p[0]})

        for v in sub.values():
            for k in sub.keys():
                docsub["{}_url".format(v)] = "{}".format(src[v]["url"])
                if "resp" == k:
                    docsub.update({"{}_resp".format(v):
                                   mkblock(src[v]["response"])})
                if "body" == k:
                    docsub.update({"{}_body".format(v):
                                   mkblock(src[v]["body"])})

                if "params" == k:
                    docsub.update({"{}_params".format(v):
                                   mkblock(src[v]["params"])})
                if "ciresp" == k:
                    docsub.update({"{}_ciresp".format(v):
                                   mkblock(src[v]["response"], 1)})

        obj.__doc__ = obj.__doc__.format(**docsub)

        return obj

    return dec


def endpoint(url, method="GET", expected_status=200):
    """endpoint - decorator to manipulate the REST-service endpoint.

    The endpoint decorator sets the endpoint and the method for the class
    to access the REST-service.
    """
    def dec(obj):
        obj.ENDPOINT = url
        obj.METHOD = method
        obj.EXPECTED_STATUS = expected_status
        return obj

    return dec


def abstractclass(cls):
    """abstractclass - class decorator.

    make sure the class is abstract and cannot be used on it's own.

    @abstractclass
    class A(object):
        def __init__(self, *args, **kwargs):
            # logic
            pass

    class B(A):
        pass

    a = A()   # results in an AssertionError
    b = B()   # works fine
    """
    setattr(cls, "_ISNEVER", cls.__bases__[0].__name__)
    origInit = cls.__dict__["__init__"]

    def wrapInit(self, *args, **kwargs):
        # when the class is instantiated we can check for bases
        # we don't want it to be the base class
        try:
            assert self.__class__.__bases__[-1].__name__ != self._ISNEVER
            origInit(self, *args, **kwargs)
        except AssertionError:
            raise TypeError("Use of abstract base class")

    # replace the original __init__
    setattr(wrapInit, "__doc__", getattr(origInit, "__doc__"))
    setattr(origInit, "__doc__", "")
    setattr(cls, "__init__", wrapInit)

    return cls


class extendargs(object):
    """'extendargs' decorator.

    Add extra arguments to the argumentlist of the constructor of the class.
    """

    def __init__(self, *loa):
        self.loa = loa

    def __call__(self, cls):
        # save parent class __init__
        origInit = cls.__bases__[0].__dict__["__init__"]

        def wrapInit(wself, *args, **kwargs):
            for extraArg in self.loa:
                if extraArg in kwargs:
                    setattr(wself, extraArg, kwargs[extraArg])
                    del kwargs[extraArg]
            origInit(wself, *args, **kwargs)
        setattr(cls, "__init__", wrapInit)

        return cls



"""Handle position endpoints."""
from .apirequest import APIRequest
from .decorators import dyndoc_insert, endpoint
from .responses.positions import responses
from abc import abstractmethod


class Positions(APIRequest):
    """Positions - abstractbase class to handle the 'positions' endpoints."""

    ENDPOINT = ""
    METHOD = "GET"

    @abstractmethod
    @dyndoc_insert(responses)
    def __init__(self, accountID, instrument=None):
        """Instantiate a Positions APIRequest instance.

        Parameters
        ----------
        accountID : string (required)
            the id of the account to perform the request on.

        instrument : string (optional)
            the instrument for the Positions request

        """
        endpoint = self.ENDPOINT.format(accountID=accountID,
                                        instrument=instrument)
        super(Positions, self).__init__(endpoint,
                                        method=self.METHOD)


@endpoint("v3/accounts/{accountID}/positions")
class PositionList(Positions):
    """PositionList.

    List all Positions for an Account. The Positions returned are for every
    instrument that has had a position during the lifetime of the Account.
    """

    @dyndoc_insert(responses)
    def __init__(self, accountID):
        """Instantiate a PositionList request.

        Parameters
        ----------
        accountID : string (required)
            id of the account to perform the request on.


        >>> import oandapyV20
        >>> import oandapyV20.endpoints.positions as positions
        >>> accountID = ...
        >>> client = oandapyV20.API(access_token=...)
        >>> r = positions.PositionList(accountID=accountID)
        >>> client.request(r)
        >>> print r.response

        Output::

            {_v3_accounts_accountID_positions_resp}

        """
        super(PositionList, self).__init__(accountID)


@endpoint("v3/accounts/{accountID}/openPositions")
class OpenPositions(Positions):
    """OpenPositions.

    List all open Positions for an Account. An open Position is a Position
    in an Account that currently has a Trade opened for it.
    """

    @dyndoc_insert(responses)
    def __init__(self, accountID):
        """Instantiate an OpenPositions request.

        Parameters
        ----------
        accountID : string (required)
            id of the account to perform the request on.


        >>> import oandapyV20
        >>> import oandapyV20.endpoints.positions as positions
        >>> accountID = ...
        >>> client = oandapyV20.API(access_token=...)
        >>> r = positions.OpenPositions(accountID=accountID)
        >>> client.request(r)
        >>> print r.response

        Output::

            {_v3_accounts_accountID_openpositions_resp}

        """
        super(OpenPositions, self).__init__(accountID)


@endpoint("v3/accounts/{accountID}/positions/{instrument}")
class PositionDetails(Positions):
    """PositionDetails.

    Get the details of a single instrument's position in an Account. The
    position may be open or not.
    """

    @dyndoc_insert(responses)
    def __init__(self, accountID, instrument):
        """Instantiate a PositionDetails request.

        Parameters
        ----------
        accountID : string (required)
            id of the account to perform the request on.

        instrument : string (required)
            id of the instrument to get the position details for.


        >>> import oandapyV20
        >>> import oandapyV20.endpoints.positions as positions
        >>> accountID = ...
        >>> instrument = ...
        >>> client = oandapyV20.API(access_token=...)
        >>> r = positions.PositionDetails(accountID=accountID, instrument)
        >>> client.request(r)
        >>> print r.response

        Output::

            {_v3_accounts_accountID_positiondetails_resp}

        """
        super(PositionDetails, self).__init__(accountID, instrument)


@endpoint("v3/accounts/{accountID}/positions/{instrument}/close", "PUT")
class PositionClose(Positions):
    """Closeout the open Position regarding instrument in an Account."""

    HEADERS = {"Content-Type": "application/json"}

    @dyndoc_insert(responses)
    def __init__(self, accountID, instrument, data):
        """Instantiate a PositionClose request.

        Parameters
        ----------
        accountID : string (required)
            id of the account to perform the request on.

        instrument : string (required)
            instrument to close partially or fully.

        data : dict (required)
            closeout specification data to send, check developer.oanda.com
            for details.


        Data body example::

            {_v3_accounts_accountID_position_close_body}


        >>> import oandapyV20
        >>> import oandapyV20.endpoints.positions as positions
        >>> accountID = ...
        >>> instrument = ...
        >>> client = oandapyV20.API(access_token=...)
        >>> data = {_v3_accounts_accountID_position_close_body}
        >>> r = positions.PositionClose(accountID=accountID,
        >>>                             instrument=instrument,
        >>>                             data=data)
        >>> client.request(r)
        >>> print r.response

        Output::

            {_v3_accounts_accountID_position_close_resp}

        """
        super(PositionClose, self).__init__(accountID, instrument)
        self.data = data



# -*- coding: utf-8 -*-
"""Transactions definitions."""

definitions = {
    "TransactionType": {
        "CREATE": "Account Create Transaction",
        "CLOSE": "Account Close Transaction",
        "REOPEN": "Account Reopen Transaction",
        "CLIENT_CONFIGURE": "Client Configuration Transaction",
        "CLIENT_CONFIGURE_REJECT": "Client Configuration Reject Transaction",
        "TRANSFER_FUNDS": "Transfer Funds Transaction",
        "TRANSFER_FUNDS_REJECT": "Transfer Funds Reject Transaction",
        "MARKET_ORDER": "Market Order Transaction",
        "MARKET_ORDER_REJECT": "Market Order Reject Transaction",
        "LIMIT_ORDER": "Limit Order Transaction",
        "LIMIT_ORDER_REJECT": "Limit Order Reject Transaction",
        "STOP_ORDER": "Stop Order Transaction",
        "STOP_ORDER_REJECT": "Stop Order Reject Transaction",
        "MARKET_IF_TOUCHED_ORDER": "Market if Touched Order Transaction",
        "MARKET_IF_TOUCHED_ORDER_REJECT": "Market if Touched Order "
                                          "Reject Transaction",
        "TAKE_PROFIT_ORDER": "Take Profit Order Transaction",
        "TAKE_PROFIT_ORDER_REJECT": "Take Profit Order Reject Transaction",
        "STOP_LOSS_ORDER": "Stop Loss Order Transaction",
        "STOP_LOSS_ORDER_REJECT": "Stop Loss Order Reject Transaction",
        "TRAILING_STOP_LOSS_ORDER": "Trailing Stop Loss Order Transaction",
        "TRAILING_STOP_LOSS_ORDER_REJECT": "Trailing Stop Loss Order "
                                           "Reject Transaction",
        "ORDER_FILL": "Order Fill Transaction",
        "ORDER_CANCEL": "Order Cancel Transaction",
        "ORDER_CANCEL_REJECT": "Order Cancel Reject Transaction",
        "ORDER_CLIENT_EXTENSIONS_MODIFY": "Order Client Extensions "
                                          "Modify Transaction",
        "ORDER_CLIENT_EXTENSIONS_MODIFY_REJECT": "Order Client Extensions "
                                          "Modify Reject Transaction",
        "TRADE_CLIENT_EXTENSIONS_MODIFY": "Trade Client Extensions "
                                          "Modify Transaction",
        "TRADE_CLIENT_EXTENSIONS_MODIFY_REJECT": "Trade Client Extensions "
                                                 "Modify Reject Transaction",
        "MARGIN_CALL_ENTER": "Margin Call Enter Transaction",
        "MARGIN_CALL_EXTEND": "Margin Call Extend Transaction",
        "MARGIN_CALL_EXIT": "Margin Call Exit Transaction",
        "DELAYED_TRADE_CLOSURE": "Delayed Trade Closure Transaction",
        "DAILY_FINANCING": "Daily Financing Transaction",
        "RESET_RESETTABLE_PL": "Reset Resettable PL Transaction",
    },
    "FundingReason": {
        "CLIENT_FUNDING": "The client has initiated a funds transfer",
        "ACCOUNT_TRANSFER": "Funds are being transfered between "
                            "two Accounts.",
        "DIVISION_MIGRATION": "Funds are being transfered as part of a "
                              "Division migration",
        "SITE_MIGRATION": "Funds are being transfered as part of a "
                          "Site migration",
        "ADJUSTMENT": "Funds are being transfered as part of an "
                      "Account adjustment"
    },
    "MarketOrderReason": {
        "CLIENT_ORDER": "The Market Order was created at the request "
                        "of a client",
        "TRADE_CLOSE": "The Market Order was created to close a Trade "
                       "at the request of a client",
        "POSITION_CLOSEOUT": "The Market Order was created to close a "
                             "Position at the request of a client",
        "MARGIN_CLOSEOUT": "The Market Order was created as part of a "
                           "Margin Closeout",
        "DELAYED_TRADE_CLOSE": "The Market Order was created to close a "
                               "trade marked for delayed closure",
    },
    "LimitOrderReason": {
        "CLIENT_ORDER": "The Limit Order was initiated at the request of a "
                        "client",
        "REPLACEMENT": "The Limit Order was initiated as a replacement for "
                       "an existing Order",
    },
    "StopOrderReason": {
        "CLIENT_ORDER": "The Stop Order was initiated at the request of a "
                        "client",
        "REPLACEMENT": "The Stop Order was initiated as a replacement for "
                       "an existing Order",
    },
    "MarketIfTouchedOrderReason": {
        "CLIENT_ORDER": "The Market-if-touched Order was initiated at the "
                        "request of a client",
        "REPLACEMENT": "The Market-if-touched Order was initiated as a "
                       "replacement for an existing Order",
    },
    "TakeProfitOrderReason": {
        "CLIENT_ORDER": "The Take Profit Order was initiated at the request "
                        "of a client",
        "REPLACEMENT": "The Take Profit Order was initiated as a replacement "
                       "for an existing Order",
        "ON_FILL": "The Take Profit Order was initiated automatically when "
                   "an Order was filled that opened a new Trade requiring "
                   "a Take Profit Order.",
    },
    "StopLossOrderReason": {
        "CLIENT_ORDER": "The Stop Loss Order was initiated at the request "
                        "of a client",
        "REPLACEMENT": "The Stop Loss Order was initiated as a replacement "
                       "for an existing Order",
        "ON_FILL": "The Stop Loss Order was initiated automatically when "
                   "an Order was filled that opened a new Trade requiring "
                   "a Stop Loss Order.",
    },
    "TrailingStopLossOrderReason": {
        "CLIENT_ORDER": "The Trailing Stop Loss Order was initiated at the "
                        "request of a client",
        "REPLACEMENT": "The Trailing Stop Loss Order was initiated as a "
                       "replacement for an existing Order",
        "ON_FILL": "The Trailing Stop Loss Order was initiated automatically "
                   "when an Order was filled that opened a new Trade "
                   "requiring a Trailing Stop Loss Order.",
    },
    "OrderFillReason": {
        "LIMIT_ORDER": "The Order filled was a Limit Order",
        "STOP_ORDER": "The Order filled was a Stop Order",
        "MARKET_IF_TOUCHED_ORDER": "The Order filled was a "
                                   "Market-if-touched Order",
        "TAKE_PROFIT_ORDER": "The Order filled was a Take Profit Order",
        "STOP_LOSS_ORDER": "The Order filled was a Stop Loss Order",
        "TRAILING_STOP_LOSS_ORDER": "The Order filled was a Trailing Stop "
                                    "Loss Order",
        "MARKET_ORDER": "The Order filled was a Market Order",
        "MARKET_ORDER_TRADE_CLOSE": "The Order filled was a Market Order "
                                    "used to explicitly close a Trade",
        "MARKET_ORDER_POSITION_CLOSEOUT": "The Order filled was a Market "
                                          "Order used to explicitly close "
                                          "a Position",
        "MARKET_ORDER_MARGIN_CLOSEOUT": "The Order filled was a Market Order "
                                        "used for a Margin Closeout",
        "MARKET_ORDER_DELAYED_TRADE_CLOSE": "The Order filled was a Market "
                                            "Order used for a delayed Trade "
                                            "close",
    },
    "OrderCancelReason": {
        "INTERNAL_SERVER_ERROR": "The Order was cancelled because at the"
                                 "time of filling, an unexpected internal "
                                 "server error occurred.",
        "ACCOUNT_LOCKED": "The Order was cancelled because at the time of "
                          "filling the account was locked.",
        "ACCOUNT_NEW_POSITIONS_LOCKED": "The order was to be filled, "
                                        "however the account is configured "
                                        "to not allow new positions to be "
                                        "created.",
        "ACCOUNT_ORDER_CREATION_LOCKED": "Filling the Order wasn’t possible "
                                         "because it required the creation "
                                         "of a dependent Order and the "
                                         "Account is locked for Order "
                                         "creation.",
        "ACCOUNT_ORDER_FILL_LOCKED": "Filling the Order was not possible "
                                     "because the Account is locked for "
                                     "filling Orders.",
        "CLIENT_REQUEST": "The Order was cancelled explicitly at the request "
                          "of the client.",
        "MIGRATION": "The Order cancelled because it is being migrated to "
                     "another account.",
        "MARKET_HALTED": "Filling the Order wasn’t possible because the "
                         "Order’s instrument was halted.",
        "LINKED_TRADE_CLOSED": "The Order is linked to an open Trade that "
                               "was closed.",
        "TIME_IN_FORCE_EXPIRED": "The time in force specified for this order "
                                 "has passed.",
        "INSUFFICIENT_MARGIN": "Filling the Order wasn’t possible because "
                               "the Account had insufficient margin.",
        "FIFO_VIOLATION": "Filling the Order would have resulted in a FIFO "
                          "violation.",
        "BOUNDS_VIOLATION": "Filling the Order would have violated the "
                            "Order’s price bound.",
        "CLIENT_REQUEST_REPLACED": "The Order was cancelled for replacement "
                                   "at the request of the client.",
        "INSUFFICIENT_LIQUIDITY": "Filling the Order wasn’t possible "
                                  "because enough liquidity available.",
        "TAKE_PROFIT_ON_FILL_GTD_TIMESTAMP_IN_PAST":
            "Filling the Order would have resulted in the creation "
            "of a Take Profit Order with a GTD time in the past.",
        "TAKE_PROFIT_ON_FILL_LOSS":
            "Filling the Order would result in the creation of a "
            "Take Profit Order that would have been filled immediately, "
            "closing the new Trade at a loss.",
        "LOSING_TAKE_PROFIT":
            "Filling the Order would result in the creation of a "
            "Take Profit Loss Order that would close the new Trade "
            "at a loss when filled.",
        "STOP_LOSS_ON_FILL_GTD_TIMESTAMP_IN_PAST":
            "Filling the Order would have resulted in the creation of a "
            "Stop Loss Order with a GTD time in the past.",
        "STOP_LOSS_ON_FILL_LOSS":
            "Filling the Order would result in the creation of a "
            "Stop Loss Order that would have been filled immediately, "
            "closing the new Trade at a loss.",
        "TRAILING_STOP_LOSS_ON_FILL_GTD_TIMESTAMP_IN_PAST":
            "Filling the Order would have resulted in the creation of a"
            "Trailing Stop Loss Order with a GTD time in the past.",
        "CLIENT_TRADE_ID_ALREADY_EXISTS":
            "Filling the Order would result in the creation of a "
            "new Open Trade with a client Trade ID already in use.",
        "POSITION_CLOSEOUT_FAILED": "Closing out a position wasn’t "
                                    "fully possible.",
        "OPEN_TRADES_ALLOWED_EXCEEDED":
            "Filling the Order would cause the maximum open trades "
            "allowed for the Account to be exceeded.",
        "PENDING_ORDERS_ALLOWED_EXCEEDED":
            "Filling the Order would have resulted in exceeding the "
            "number of pending Orders allowed for the Account.",
        "TAKE_PROFIT_ON_FILL_CLIENT_ORDER_ID_ALREADY_EXISTS":
            "Filling the Order would have resulted in the creation of "
            "a Take Profit Order with a client Order ID that is already "
            "in use.",
        "STOP_LOSS_ON_FILL_CLIENT_ORDER_ID_ALREADY_EXISTS":
            "Filling the Order would have resulted in the creation of a "
            "Stop Loss Order with a client Order ID that is already in use.",
        "TRAILING_STOP_LOSS_ON_FILL_CLIENT_ORDER_ID_ALREADY_EXISTS":
            "Filling the Order would have resulted in the creation of a "
            "Trailing Stop Loss Order with a client Order ID that is "
            "already in use.",
        "POSITION_SIZE_EXCEEDED":
            "Filling the Order would have resulted in the "
            "Account’s maximum position size limit being exceeded "
            "for the Order’s instrument.",
    },
    "MarketOrderMarginCloseoutReason": {
        "MARGIN_CHECK_VIOLATION": "Trade closures resulted from violating "
                                  "OANDA’s margin policy",
        "REGULATORY_MARGIN_CALL_VIOLATION":
            "Trade closures came from a margin closeout event resulting "
            "from regulatory conditions placed on the Account’s margin call"
    },
    "TransactionRejectReason": {
        "INTERNAL_SERVER_ERROR": "An unexpected internal server error has "
                                 "occurred",
        "INSTRUMENT_PRICE_UNKNOWN": "The system was unable to determine the "
                                    "current price for the Order’s instrument",
        "ACCOUNT_NOT_ACTIVE": "The Account is not active",
        "ACCOUNT_LOCKED": "The Account is locked",
        "ACCOUNT_ORDER_CREATION_LOCKED": "The Account is locked for Order "
                                         "creation",
        "ACCOUNT_CONFIGURATION_LOCKED": "The Account is locked for "
                                        "configuration",
        "ACCOUNT_DEPOSIT_LOCKED": "The Account is locked for deposits",
        "ACCOUNT_WITHDRAWAL_LOCKED": "The Account is locked for withdrawals",
        "ACCOUNT_ORDER_CANCEL_LOCKED": "The Account is locked for Order "
                                       "cancellation",
        "INSTRUMENT_NOT_TRADEABLE": "The instrument specified is not "
                                    "tradeable by the Account",
        "PENDING_ORDERS_ALLOWED_EXCEEDED": "Creating the Order would result "
                                           "in the maximum number of allowed "
                                           "pending Orders being exceeded",
        "ORDER_ID_UNSPECIFIED": "Neither the Order ID nor client Order ID "
                                "are specified",
        "ORDER_DOESNT_EXIST": "The Order specified does not exist",
        "ORDER_IDENTIFIER_INCONSISTENCY": "The Order ID and client Order ID "
                                          "specified do not identify the "
                                          "same Order",
        "TRADE_ID_UNSPECIFIED": "Neither the Trade ID nor client Trade ID "
                                "are specified",
        "TRADE_DOESNT_EXIST": "The Trade specified does not exist",
        "TRADE_IDENTIFIER_INCONSISTENCY": "The Trade ID and client Trade ID "
                                          "specified do not identify the "
                                          "same Trade",
        "INSTRUMENT_MISSING": "Order instrument has not been specified",
        "INSTRUMENT_UNKNOWN": "The instrument specified is unknown",
        "UNITS_MISSING": "Order units have not been not specified",
        "UNITS_INVALID": "Order units specified are invalid",
        "UNITS_PRECISION_EXCEEDED": "The units specified contain more "
                                    "precision than is allowed for the "
                                    "Order’s instrument",
        "UNITS_LIMIT_EXCEEDED": "The units specified exceeds the maximum "
                                "number of units allowed",
        "UNITS_MIMIMUM_NOT_MET": "The units specified is less than the "
                                 "minimum number of units required",
        "PRICE_MISSING": "The price has not been specified",
        "PRICE_INVALID": "The price specifed is invalid",
        "PRICE_PRECISION_EXCEEDED": "The price specified contains more "
                                    "precision than is allowed for the "
                                    "instrument",
        "PRICE_DISTANCE_MISSING": "The price distance has not been specified",
        "PRICE_DISTANCE_INVALID": "The price distance specifed is invalid",
        "PRICE_DISTANCE_PRECISION_EXCEEDED": "The price distance specified "
                                             "contains more precision than is "
                                             "allowed for the instrument",
        "PRICE_DISTANCE_MAXIMUM_EXCEEDED": "The price distance exceeds that "
                                           "maximum allowed amount",
        "PRICE_DISTANCE_MINIMUM_NOT_MET": "The price distance does not meet "
                                          "the minimum allowed amount",
        "TIME_IN_FORCE_MISSING": "The TimeInForce field has not been "
                                 "specified",
        "TIME_IN_FORCE_INVALID": "The TimeInForce specified is invalid",
        "TIME_IN_FORCE_GTD_TIMESTAMP_MISSING": "The TimeInForce is GTD but no "
                                               "GTD timestamp is provided",
        "TIME_IN_FORCE_GTD_TIMESTAMP_IN_PAST": "The TimeInForce is GTD but "
                                               "the GTD timestamp is in the "
                                               "past",
        "PRICE_BOUND_INVALID": "The price bound specified is invalid",
        "PRICE_BOUND_PRECISION_EXCEEDED": "The price bound specified contains "
                                          "more precision than is allowed for "
                                          "the Order’s instrument",
        "ORDERS_ON_FILL_DUPLICATE_CLIENT_ORDER_IDS":
            "Multiple Orders on fill share the same client Order ID",
        "TRADE_ON_FILL_CLIENT_EXTENSIONS_NOT_SUPPORTED":
            "The Order does not support Trade on fill client extensions "
            "because it cannot create a new Trade",
        "CLIENT_ORDER_ID_INVALID": "The client Order ID specified is invalid",
        "CLIENT_ORDER_ID_ALREADY_EXISTS":
            "The client Order ID specified is already assigned to another "
            "pending Order",
        "CLIENT_ORDER_TAG_INVALID":
            "The client Order tag specified is invalid",
        "CLIENT_ORDER_COMMENT_INVALID":
            "The client Order comment specified is invalid",
        "CLIENT_TRADE_ID_INVALID": "The client Trade ID specified is invalid",
        "CLIENT_TRADE_ID_ALREADY_EXISTS":
            "The client Trade ID specifed is already assigned to another "
            "open Trade",
        "CLIENT_TRADE_TAG_INVALID":
            "The client Trade tag specified is invalid",
        "CLIENT_TRADE_COMMENT_INVALID": "The client Trade comment is invalid",
        "ORDER_FILL_POSITION_ACTION_MISSING":
            "The OrderFillPositionAction field has not been specified",
        "ORDER_FILL_POSITION_ACTION_INVALID":
            "The OrderFillPositionAction specified is invalid",
        "TRIGGER_CONDITION_MISSING":
            "The TriggerCondition field has not been specified",
        "TRIGGER_CONDITION_INVALID":
            "The TriggerCondition specified is invalid",
        "ORDER_PARTIAL_FILL_OPTION_MISSING":
            "The OrderFillPositionAction field has not been specified",
        "ORDER_PARTIAL_FILL_OPTION_INVALID":
            "The OrderFillPositionAction specified is invalid.",
        "INVALID_REISSUE_IMMEDIATE_PARTIAL_FILL":
            "When attempting to reissue an order (currently only a "
            "MarketIfTouched) that was immediately partially filled, it is "
            "not possible to create a correct pending Order.",
        "TAKE_PROFIT_ORDER_ALREADY_EXISTS":
            "A Take Profit Order for the specified Trade already exists",
        "TAKE_PROFIT_ON_FILL_PRICE_MISSING":
            "The Take Profit on fill specified does not provide a price",
        "TAKE_PROFIT_ON_FILL_PRICE_INVALID":
            "The Take Profit on fill specified contains an invalid price",
        "TAKE_PROFIT_ON_FILL_PRICE_PRECISION_EXCEEDED":
            "The Take Profit on fill specified contains a price with more "
            "precision than is allowed by the Order’s instrument",
        "TAKE_PROFIT_ON_FILL_TIME_IN_FORCE_MISSING":
            "The Take Profit on fill specified does not provide a TimeInForce",
        "TAKE_PROFIT_ON_FILL_TIME_IN_FORCE_INVALID":
            "The Take Profit on fill specifies an invalid TimeInForce",
        "TAKE_PROFIT_ON_FILL_GTD_TIMESTAMP_MISSING":
            "The Take Profit on fill specifies a GTD TimeInForce but does "
            "not provide a GTD timestamp",
        "TAKE_PROFIT_ON_FILL_GTD_TIMESTAMP_IN_PAST":
            "The Take Profit on fill specifies a GTD timestamp that is in "
            "the past",
        "TAKE_PROFIT_ON_FILL_CLIENT_ORDER_ID_INVALID":
            "The Take Profit on fill client Order ID specified is invalid",
        "TAKE_PROFIT_ON_FILL_CLIENT_ORDER_TAG_INVALID":
            "The Take Profit on fill client Order tag specified is invalid",
        "TAKE_PROFIT_ON_FILL_CLIENT_ORDER_COMMENT_INVALID":
            "The Take Profit on fill client Order comment specified is "
            "invalid",
        "TAKE_PROFIT_ON_FILL_TRIGGER_CONDITION_MISSING":
            "The Take Profit on fill specified does not provide a "
            "TriggerCondition",
        "TAKE_PROFIT_ON_FILL_TRIGGER_CONDITION_INVALID":
            "The Take Profit on fill specifies an invalid TriggerCondition",
        "STOP_LOSS_ORDER_ALREADY_EXISTS":
            "A Stop Loss Order for the specified Trade already exists",
        "STOP_LOSS_ON_FILL_PRICE_MISSING":
            "The Stop Loss on fill specified does not provide a price",
        "STOP_LOSS_ON_FILL_PRICE_INVALID":
            "The Stop Loss on fill specifies an invalid price",
        "STOP_LOSS_ON_FILL_PRICE_PRECISION_EXCEEDED":
            "The Stop Loss on fill specifies a price with more precision "
            "than is allowed by the Order’s instrument",
        "STOP_LOSS_ON_FILL_TIME_IN_FORCE_MISSING":
            "The Stop Loss on fill specified does not provide a TimeInForce",
        "STOP_LOSS_ON_FILL_TIME_IN_FORCE_INVALID":
            "The Stop Loss on fill specifies an invalid TimeInForce",
        "STOP_LOSS_ON_FILL_GTD_TIMESTAMP_MISSING":
            "The Stop Loss on fill specifies a GTD TimeInForce but does "
            "not provide a GTD timestamp",
        "STOP_LOSS_ON_FILL_GTD_TIMESTAMP_IN_PAST":
            "The Stop Loss on fill specifies a GTD timestamp that is in "
            "the past",
        "STOP_LOSS_ON_FILL_CLIENT_ORDER_ID_INVALID":
            "The Stop Loss on fill client Order ID specified is invalid",
        "STOP_LOSS_ON_FILL_CLIENT_ORDER_TAG_INVALID":
            "The Stop Loss on fill client Order tag specified is invalid",
        "STOP_LOSS_ON_FILL_CLIENT_ORDER_COMMENT_INVALID":
            "The Stop Loss on fill client Order comment specified is invalid",
        "STOP_LOSS_ON_FILL_TRIGGER_CONDITION_MISSING":
            "The Stop Loss on fill specified does not provide a "
            "TriggerCondition",
        "STOP_LOSS_ON_FILL_TRIGGER_CONDITION_INVALID":
            "The Stop Loss on fill specifies an invalid TriggerCondition",
        "TRAILING_STOP_LOSS_ORDER_ALREADY_EXISTS":
            "A Trailing Stop Loss Order for the specified Trade already "
            "exists",
        "TRAILING_STOP_LOSS_ON_FILL_PRICE_DISTANCE_MISSING":
            "The Trailing Stop Loss on fill specified does not provide a "
            "distance",
        "TRAILING_STOP_LOSS_ON_FILL_PRICE_DISTANCE_INVALID":
            "The Trailing Stop Loss on fill distance is invalid",
        "TRAILING_STOP_LOSS_ON_FILL_PRICE_DISTANCE_PRECISION_EXCEEDED":
            "The Trailing Stop Loss on fill distance contains more precision "
            "than is allowed by the instrument",
        "TRAILING_STOP_LOSS_ON_FILL_PRICE_DISTANCE_MAXIMUM_EXCEEDED":
            "The Trailing Stop Loss on fill price distance exceeds the "
            "maximum allowed amount",
        "TRAILING_STOP_LOSS_ON_FILL_PRICE_DISTANCE_MINIMUM_NOT_MET":
            "The Trailing Stop Loss on fill price distance does not meet "
            "the minimum allowed amount",
        "TRAILING_STOP_LOSS_ON_FILL_TIME_IN_FORCE_MISSING":
            "The Trailing Stop Loss on fill specified does not provide a "
            "TimeInForce",
        "TRAILING_STOP_LOSS_ON_FILL_TIME_IN_FORCE_INVALID":
            "The Trailing Stop Loss on fill specifies an invalid TimeInForce",
        "TRAILING_STOP_LOSS_ON_FILL_GTD_TIMESTAMP_MISSING":
            "The Trailing Stop Loss on fill TimeInForce is specified as GTD "
            "but no GTD timestamp is provided",
        "TRAILING_STOP_LOSS_ON_FILL_GTD_TIMESTAMP_IN_PAST":
            "The Trailing Stop Loss on fill GTD timestamp is in the past",
        "TRAILING_STOP_LOSS_ON_FILL_CLIENT_ORDER_ID_INVALID":
            "The Trailing Stop Loss on fill client Order ID specified is "
            "invalid",
        "TRAILING_STOP_LOSS_ON_FILL_CLIENT_ORDER_TAG_INVALID":
            "The Trailing Stop Loss on fill client Order tag specified is "
            "invalid",
        "TRAILING_STOP_LOSS_ON_FILL_CLIENT_ORDER_COMMENT_INVALID":
            "The Trailing Stop Loss on fill client Order comment specified "
            "is invalid",
        "TRAILING_STOP_LOSS_ORDERS_NOT_SUPPORTED":
            "A client attempted to create either a Trailing Stop Loss order "
            "or an order with a Trailing Stop Loss On Fill specified, which "
            "may not yet be supported.",
        "TRAILING_STOP_LOSS_ON_FILL_TRIGGER_CONDITION_MISSING":
            "The Trailing Stop Loss on fill specified does not provide a "
            "TriggerCondition",
        "TRAILING_STOP_LOSS_ON_FILL_TRIGGER_CONDITION_INVALID":
            "The Tailing Stop Loss on fill specifies an invalid "
            "TriggerCondition",
        "CLOSE_TRADE_TYPE_MISSING": "The request to close a Trade does not "
                                    "specify a full or partial close",
        "CLOSE_TRADE_PARTIAL_UNITS_MISSING":
            "The request to close a Trade partially did not specify the "
            "number of units to close",
        "CLOSE_TRADE_UNITS_EXCEED_TRADE_SIZE":
            "The request to partially close a Trade specifies a number of "
            "units that exceeds the current size of the given Trade",
        "CLOSEOUT_POSITION_DOESNT_EXIST":
            "The Position requested to be closed out does not exist",
        "CLOSEOUT_POSITION_INCOMPLETE_SPECIFICATION":
            "The request to closeout a Position was specified incompletely",
        "CLOSEOUT_POSITION_UNITS_EXCEED_POSITION_SIZE":
            "A partial Position closeout request specifies a number of units "
            "that exceeds the current Position",
        "CLOSEOUT_POSITION_REJECT":
            "The request to closeout a Position could not be fully satisfied",
        "CLOSEOUT_POSITION_PARTIAL_UNITS_MISSING":
            "The request to partially closeout a Position did not specify "
            "the number of units to close.",
        "MARKUP_GROUP_ID_INVALID": "The markup group ID provided is invalid",
        "POSITION_AGGREGATION_MODE_INVALID":
            "The PositionAggregationMode provided is not supported/valid.",
        "ADMIN_CONFIGURE_DATA_MISSING": "No configuration parameters provided",
        "MARGIN_RATE_INVALID": "The margin rate provided is invalid",
        "MARGIN_RATE_WOULD_TRIGGER_CLOSEOUT": "The margin rate provided "
                                              "would cause an immediate "
                                              "margin closeout",
        "ALIAS_INVALID": "The account alias string provided is invalid",
        "CLIENT_CONFIGURE_DATA_MISSING":
            "No configuration parameters provided",
        "MARGIN_RATE_WOULD_TRIGGER_MARGIN_CALL":
            "The margin rate provided would cause the Account to enter "
            "a margin call state.",
        "AMOUNT_INVALID": "Funding is not possible because the requested "
                          "transfer amount is invalid",
        "INSUFFICIENT_FUNDS": "The Account does not have sufficient "
                              "balance to complete the funding request",
        "AMOUNT_MISSING": "Funding amount has not been specified",
        "FUNDING_REASON_MISSING": "Funding reason has not been specified",
        "CLIENT_EXTENSIONS_DATA_MISSING":
            "Neither Order nor Trade on Fill client extensions were "
            "provided for modification",
        "REPLACING_ORDER_INVALID":
            "The Order to be replaced has a different type than the "
            "replacing Order.",
        "REPLACING_TRADE_ID_INVALID":
            "The replacing Order refers to a different Trade than the "
            "Order that is being replaced.",
     }
}



# -*- coding: utf-8 -*-
"""baserequest."""

import json
import six
from abc import ABCMeta, abstractmethod


@six.add_metaclass(ABCMeta)
class BaseRequest(object):
    """baseclass for request classes."""

    @abstractmethod
    def __init__(self):
        self._data = dict()

    def __repr__(self):
        return json.dumps(self.__dict__)

    @property
    def data(self):
        """data - return the JSON body.

        The data property returns a dict representing the
        JSON-body needed for the API-request. All values that are
        not set will be left out
        """
        d = dict()
        for k, v in self._data.items():
            # skip unset properties
            if v is None:
                continue

            d.update({k: v})

        return d

    def toJSON(self):
        return json.dumps(self,
                          default=lambda o: o.__dict__,
                          sort_keys=True,
                          indent=4)



# -*- coding: utf-8 -*-
"""Instruments Definitions."""

definitions = {
    "PriceComponents": {
        "A": "Ask",
        "B": "Bid",
        "M": "Mid",
    },
    "CandlestickGranularity": {
        "S5": "5 second candlesticks, minute alignment",
        "S10": "10 second candlesticks, minute alignment",
        "S15": "15 second candlesticks, minute alignment",
        "S30": "30 second candlesticks, minute alignment",
        "M1": "1 minute candlesticks, minute alignment",
        "M2": "2 minute candlesticks, hour alignment",
        "M4": "4 minute candlesticks, hour alignment",
        "M5": "5 minute candlesticks, hour alignment",
        "M10": "10 minute candlesticks, hour alignment",
        "M15": "15 minute candlesticks, hour alignment",
        "M30": "30 minute candlesticks, hour alignment",
        "H1": "1 hour candlesticks, hour alignment",
        "H2": "1 hour candlesticks, day alignment",
        "H3": "3 hour candlesticks, day alignment",
        "H4": "4 hour candlesticks, day alignment",
        "H6": "6 hour candlesticks, day alignment",
        "H8": "8 hour candlesticks, day alignment",
        "H12": "12 hour candlesticks, day alignment",
        "D": "1 day candlesticks, day alignment",
        "W": "1 week candlesticks, aligned to start of week",
        "M": "1 month candlesticks, aligned to first day of the month",
    },
    "WeeklyAlignment": {
        "Monday": "Monday",
        "Tuesday": "Tuesday",
        "Wednesday": "Wednesday",
        "Thursday": "Thursday",
        "Friday": "Friday",
        "Saturday": "Saturday",
        "Sunday": "Sunday",
    },
}



from setuptools import setup, find_packages
import os
import sys
import re
import shutil


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


def read_all(f):
    with open(f) as I:
        return I.read()

requirements = map(str.strip, open("requirements.txt").readlines())

version = get_version('oandapyV20')

if sys.argv[-1] == 'publish':
    print("You probably want to also tag the version now:")
    print("  git tag -a %s -m 'version %s'" % (version, version))
    print("  git push --tags")
    shutil.rmtree('dist')
    shutil.rmtree('build')
    shutil.rmtree('oandapyV20.egg-info')
    sys.exit()

setup(name='oandapyV20',
      version=version,
      description="Python wrapper for the OANDA REST-V20 API",
      long_description=read_all("README.rst"),
      classifiers=[
            'Programming Language :: Python',
            'License :: OSI Approved :: MIT License',
            'Intended Audience :: Developers',
            'Intended Audience :: Financial and Insurance Industry',
            'Operating System :: OS Independent',
            'Development Status :: 3 - Alpha',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3.3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
      ],  # Get from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='OANDA FOREX/CFD wrapper REST-V20 API',
      author='F. Brekeveld',
      author_email='f.brekeveld@gmail.com',
      url='http://github.com/hootnot/oanda-api-v20',
      license='MIT',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      test_suite="tests",
      include_package_data=True,
      zip_safe=False,
      install_requires=requirements,
      entry_points="""
      # -*- Entry points: -*-
      """,
      )
